# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions for ontology generation.
"""
import os
import sys
import re
import json
import fnmatch

from pathlib import Path
from pydoc import locate
from importlib import util as import_util
from typing import Optional, List, Tuple

from distutils.file_util import copy_file
from distutils.dir_util import mkpath
from distutils.errors import DistutilsFileError
from distutils import log

import jsonschema

AUTO_GEN_SIGNATURE = "***automatically_generated***"


def copytree(
    src,
    dst,
    ignore_pattern_if_file_exists="*",
    preserve_mode=True,
    preserve_times=False,
    preserve_symlinks=False,
    update=False,
    verbose=1,
    dry_run=0,
):
    """
    A slightly modified version of `distutils.dir_util.copytree`, with the
    added parameter `ignore_pattern_if_file_exists` that ignores files if it
    matches the given pattern and already exists in the destination directory.
    By default, ignores all existing destination files.
    """

    if not dry_run and not os.path.isdir(src):
        raise DistutilsFileError("cannot copy tree '%s': not a directory" % src)
    try:
        names = os.listdir(src)
    except OSError as e:
        if dry_run:
            names = []
        else:
            raise DistutilsFileError(
                "error listing files in '%s': %s" % (src, e.strerror)
            ) from e

    if not dry_run:
        mkpath(dst, verbose=verbose)

    outputs = []
    outputs_ignored = []

    for n in names:
        src_name = os.path.join(src, n)
        dst_name = os.path.join(dst, n)

        if n.startswith(".nfs"):
            # skip NFS rename files
            continue

        if preserve_symlinks and os.path.islink(src_name):
            link_dest = os.readlink(src_name)
            if verbose >= 1:
                log.info("linking %s -> %s", dst_name, link_dest)
            if not dry_run:
                os.symlink(link_dest, dst_name)
            outputs.append(dst_name)

        elif os.path.isdir(src_name):
            part_outputs, part_outputs_ignored = copytree(
                src_name,
                dst_name,
                ignore_pattern_if_file_exists,
                preserve_mode,
                preserve_times,
                preserve_symlinks,
                update,
                verbose=verbose,
                dry_run=dry_run,
            )
            outputs.extend(part_outputs)
            outputs_ignored.extend(part_outputs_ignored)

        elif not (
            fnmatch.fnmatch(dst_name, ignore_pattern_if_file_exists)
            and os.path.exists(dst_name)
        ):
            copy_file(
                src_name,
                dst_name,
                preserve_mode,
                preserve_times,
                update,
                verbose=verbose,
                dry_run=dry_run,
            )
            outputs.append(dst_name)

        else:
            outputs_ignored.append(dst_name)

    return outputs, outputs_ignored


def get_user_objects_from_module(
    module_str: str, custom_dirs: Optional[str] = None
):
    """
    Args:
        module_str: Module in the form of string, package.module.
        custom_dirs: custom directories to search from if `module_str` is not a
          part of imported modules.

    Returns: A list of objects present in the module `module_str`,
             None if module not found

    """
    module = locate(module_str)
    if module is not None and hasattr(module, "__all__"):
        return module.__all__  # type: ignore
    objects: List[str] = []
    if custom_dirs is not None:
        module_file = module_str.replace(".", "/") + ".py"
        for dir_ in custom_dirs:
            filepath = os.path.join(dir_, module_file)
            try:
                spec = import_util.spec_from_file_location(module_str, filepath)
                module = import_util.module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(module)  # type: ignore
                objects = module.__all__
            except (FileNotFoundError, AttributeError):
                continue
    return objects


def search_in_dirs(file, dirs_paths):
    """
    Args:
        file: File to be searched for
        dirs_paths: Directory Paths in which the `file` is to be searched

    Returns: Resolved filename if the `file` is found in `dir_paths`,
    else `None`

    """
    abs_file = file
    for _dir in dirs_paths:
        abs_dir = os.path.abspath(_dir)
        if not os.path.isabs(file):
            abs_file = os.path.join(abs_dir, file)
        abs_file = os.path.normpath(abs_file)
        if os.path.exists(abs_file):
            return abs_file
    return None


def get_module_path(module: str):
    module_spec = import_util.find_spec(module)
    return module_spec.origin if module_spec is not None else None


def get_top_level_dirs(path: Optional[str]):
    """
    Args:
        path: Path for which the directories at depth==1 are to be returned
    Returns:
        Directories at depth==1 for `path`.
    """
    if path is None or not os.path.exists(path):
        return []
    return [
        item
        for item in os.listdir(path)
        if os.path.isdir(os.path.join(path, item))
    ]


def split_file_path(path: str):
    """
    Args:
        path: Path to be split

    Returns: list containing path components
    Examples:
        >>> split_file_path('forte/data/ontology/file.py')
        ['forte', 'data', 'ontology', 'file.py']
        >>> split_file_path('/home/file.py')
        ['', 'home', 'file.py']
    """
    path_split = []
    prev_dir, curr_dir = None, (str(Path(path)), "")
    while prev_dir != curr_dir:
        prev_dir = curr_dir
        if curr_dir[-1].strip():
            path_split.append(curr_dir[-1])
        curr_dir = os.path.split(curr_dir[0])
    path_split += [""] if path.startswith("/") else []
    return path_split[::-1]


def validate_json_schema(input_filepath: str):
    """
    Validates the input json schema using validation meta-schema provided in
    `validation_filepath.json` according to the specification in
    `http://json-schema.org`.
    If the tested json is not valid, a `jsonschema.exceptions.ValidationError`
    is thrown.
    Args:
        input_filepath: Filepath of the json schema to be validated
    """
    validation_file_path = os.path.normpath(
        os.path.join(os.path.dirname(__file__), "validation_schema.json")
    )
    with open(
        validation_file_path, "r", encoding="utf-8"
    ) as validation_json_file:
        validation_schema = json.loads(validation_json_file.read())
    with open(input_filepath, "r", encoding="utf-8") as input_json_file:
        input_schema = json.loads(input_json_file.read())
    jsonschema.Draft6Validator(validation_schema).validate(input_schema)


def get_python_version() -> Tuple[int, int]:
    """
    :return: Python major and minor version at runtime
    """
    version_info = sys.version_info
    return version_info[0], version_info[1]


def get_schema_from_ontology(
    imported_onto_file: Optional[str], delimiters: List[str]
) -> str:
    if imported_onto_file is None:
        raise FileNotFoundError
    with open(imported_onto_file, "r", encoding="utf-8") as imported_onto:
        regex = "|".join(map(re.escape, delimiters))
        reqd_line = imported_onto.readlines()[1]
        installed_json_file = list(filter(None, re.split(regex, reqd_line)))[0]
    return installed_json_file


def get_parent_path(file_path: str, level: int = 1):
    relative_path = os.path.join(file_path, *([os.pardir] * level))
    return os.path.normpath(relative_path)


def get_installed_forte_dir():
    init_path = get_module_path("forte")
    return get_parent_path(init_path, 2) if init_path is not None else None


def get_current_forte_dir():
    return get_parent_path(__file__, 4)


def get_generated_files_in_dir(path):
    def is_generated(file_path):
        with open(file_path, "r", encoding="utf-8") as f:
            lines = f.readlines()
            return len(lines) > 0 and lines[0] == f"# {AUTO_GEN_SIGNATURE}\n"

    ext_files = []
    for root, _, files in os.walk(path):
        for file in files:
            path = os.path.join(root, file)
            if is_generated(path):
                ext_files.append(path)
    return ext_files



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions
"""
import difflib
from functools import wraps
from inspect import getfullargspec
from pydoc import locate
from typing import Dict, List, Optional, get_type_hints, Tuple

from sortedcontainers import SortedList
from typing_inspect import is_union_type, get_origin

__all__ = [
    "get_full_module_name",
    "get_class_name",
    "get_class",
    "get_qual_name",
    "create_class_with_kwargs",
    "check_type",
    "DiffAligner",
    "create_import_error_msg",
]


def get_full_module_name(o, lower: bool = False) -> str:
    r"""Returns the full module and class name of an object ``o``. For example,
    for our :class:`OntonotesReader`, returns
    :class:'forte.data.readers.ontonotes_reader.OntonotesReader'.

    Args:
        o (object): An object class.
        lower (bool): Whether to lowercase the full module and class name.

    Returns:
         The full module and class name.
    """
    if not isinstance(o, type):
        o = o.__class__
    module = o.__module__
    if module is None or module == str.__class__.__module__:
        return o.__name__
    name = module + "." + o.__name__
    if lower:
        return name.lower()
    else:
        return name


def get_class_name(o, lower: bool = False) -> str:
    r"""Returns the class name of an object ``o``.

    Args:
        o (object): An object class.
        lower (bool): Whether to lowercase the class name.

    Returns:
         The class name.
    """
    if not isinstance(o, type):
        o = o.__class__
    if lower:
        return o.__name__.lower()
    else:
        return o.__name__


def get_class(class_name: str, module_paths: Optional[List[str]] = None):
    r"""Returns the class based on class name.

    Args:
        class_name (str): Name or full path to the class.
        module_paths (list): Paths to candidate modules to search for the
            class. This is used if the class cannot be located solely based on
            ``class_name``. The first module in the list that contains the class
            is used.

    Returns:
        The target class.

    Raises:
        ValueError: If class is not found based on :attr:`class_name` and
            :attr:`module_paths`.
    """
    class_ = locate(class_name)
    if (class_ is None) and (module_paths is not None):
        for module_path in module_paths:
            class_ = locate(".".join([module_path, class_name]))
            if class_ is not None:
                break

    if class_ is None:
        if module_paths:
            raise ValueError(
                "Class not found in {}: {}".format(module_paths, class_name)
            )
        else:
            raise ValueError("Class not found in {}".format(class_name))

    return class_


def get_qual_name(o: object, lower: bool = False) -> str:
    r"""Returns the qualified name of an object ``o``.

    Args:
        o: An object class.
        lower: Whether to lowercase the qualified class name.

    Returns:
         The qualified class name.
    """
    if not isinstance(o, type):
        o = o.__class__
    if lower:
        return o.__qualname__.lower()
    else:
        return o.__qualname__


def create_class_with_kwargs(class_name: str, class_args: Dict):
    r"""Create class with the given arguments.

    Args:
        class_name: Class name.
        class_args: Class arguments.

    Returns:
        An object with class of type `class_name`.
    """
    cls = get_class(class_name)
    if not class_args:
        class_args = {}
    obj = cls(**class_args)

    return obj


def check_type(obj, tp):
    if is_union_type(tp):
        return any(check_type(obj, a) for a in tp.__args__)
    else:
        origin = get_origin(tp)
        if origin is None or origin == tp:
            return isinstance(obj, tp)
        else:
            return check_type(obj, origin)


def validate_input(func, **kwargs):
    hints = get_type_hints(func)

    # iterate all type hints
    for attr_name, attr_type in hints.items():
        if attr_name == "return":
            continue

        if not isinstance(kwargs[attr_name], attr_type):
            raise TypeError(
                f"{attr_name} should be of type {attr_type}, "
                f"got type {type(kwargs[attr_name])}"
            )


def type_check(func):
    @wraps(func)
    def wrapped_decorator(*args, **kwargs):
        # translate *args into **kwargs
        func_args = getfullargspec(func)[0]
        kwargs.update(dict(zip(func_args, args)))
        validate_input(func, **kwargs)
        return func(**kwargs)

    return wrapped_decorator


class DiffAligner:
    def __init__(self):
        self.__matcher = difflib.SequenceMatcher(autojunk=False)
        self.__begin_mapper = DiffAligner.OffsetMapper()
        self.__end_mapper = DiffAligner.OffsetMapper()

    class OffsetMapper:
        def __init__(self):
            self.ref_points = SortedList()
            self.adjustments = []

        def set_ref_points(self, p: int, adjustment: int):
            self.ref_points.add(p)
            self.adjustments.append(adjustment)

        def trans_offset(self, origin: int) -> int:
            """
            Translate the offset based on the reference points. If there is no
            translation found, will return None.

            Args:
                origin: The original offset

            Returns:
                The translated offset.

            """
            idx = self.ref_points.bisect_left(origin)
            if idx >= len(self.adjustments):
                return origin + self.adjustments[-1]
            return origin + self.adjustments[idx]

        def clear(self):
            self.ref_points.clear()
            self.adjustments.clear()

    def align_with_segments(
        self, text: str, segments: List[str]
    ) -> List[Optional[Tuple[int, int]]]:
        """
        Provided a text sequence `text`, and a list of text `segments`,
        this function will try to align the `text` with the `segments`,
        and return the best guess on how each segment is mapped to the
        original text. The return value will contain the offset of each
        segment, after mapping on the original text. The guess process is
        based on the "SequenceMatcher" in difflib of Python.

        Calling this function will create a mapping stored inisde the class.

        This is useful as a approximated way to map a text segmentation that
        does not contain offset information to the original text. For example,

        >>> aligner = DiffAligner()
        >>> aligner.align_with_segments("aa bb   cc", ["aa", "bb", "cc"])
        [(0, 2), (3, 5), (6, 10)]

        Args:
            text: The original text.
            segments: The list of segments to be mapped.

        Returns:
            The list of mapped offsets for each segment.
        """
        aligned_spans: List[Optional[Tuple[int, int]]] = []
        segment_spans: List[Tuple[int, int]] = []

        # Construct the text using the segments, and store position and
        # length of each element.
        segment_text: str = ""
        offset = 0
        for s in segments:
            segment_text += s
            segment_text += " "
            segment_spans.append((offset, offset + len(s)))
            offset += len(s) + 1

        self.build_alignment(text.strip(), segment_text)

        for ss in segment_spans:
            b, e = self.trans_span(ss)
            if b >= e or b >= len(text) or e > len(text):
                aligned_spans.append(None)
            else:
                aligned_spans.append((b, e))

        return aligned_spans

    def trans_span(self, span: Tuple[int, int]) -> Tuple[int, int]:
        """
        Translate the provided span based on the alignment computed. Need to
        run after `build_alignment`.

        Args:
            span: The span to be translated.

        Returns:
            The translated span.
        """
        return (
            self.__begin_mapper.trans_offset(span[0]),
            self.__end_mapper.trans_offset(span[1]),
        )

    def clear_alignment(self):
        self.__begin_mapper.clear()
        self.__end_mapper.clear()

    def build_alignment(self, text: str, text_to_align: str):
        # Compute character level mappings.
        self.__matcher.set_seqs(text, text_to_align)
        self.clear_alignment()

        for _, i1, i2, j1, j2 in self._get_opcodes():
            self.__begin_mapper.set_ref_points(j1, i1 - j1)
            self.__end_mapper.set_ref_points(j2, i2 - j2)

    def _get_opcodes(self):
        yield from self.__matcher.get_opcodes()


def create_import_error_msg(
    extra_module: str,
    forte_module: str,
    component_name: str,
    pip_installable: bool = True,
):
    """
    Create an error message for importing package extra required by a forte
    module.


    Args:
        extra_module: module name should be installed by pip.
        forte_module: forte module User should install by
            ``pip install forte[`forte_module`]`` to install all
            extra packages for using the forte module.
        component_name: the forte component that needs the module.

    """
    install_msg = (
        f" `{extra_module}` is not installed correctly."
        + f" Consider install {extra_module}"
    )
    pip_msg = f" via `pip install {extra_module}` "
    refer_msg = (
        f" or refer to extra requirement for {component_name}"
        + " at https://github.com/asyml/forte#installation"
        + f" for more information about installing {forte_module}. "
    )
    if pip_installable:
        error_msg = install_msg + pip_msg + refer_msg
    else:
        error_msg = install_msg + refer_msg
    return error_msg



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gzip
import logging
import os
from abc import ABC, abstractmethod
from typing import Iterator, List, Any, Optional

from smart_open import open

from forte.common import Resources
from forte.common.configuration import Config
from forte.common.exception import ProcessExecutionException
from forte.data.base_reader import PackReader, MultiPackReader
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack

__all__ = [
    "RawDataDeserializeReader",
    "RecursiveDirectoryDeserializeReader",
    "DirPackReader",
    "MultiPackDirectoryReader",
    "MultiPackDeserializerBase",
    "SinglePackReader",
]


class BaseDeserializeReader(PackReader, ABC):
    def _cache_key_function(self, _) -> str:
        return "cached_string_file"

    def _parse_pack(self, data_source: Any) -> Iterator[DataPack]:
        if data_source is None:
            raise ProcessExecutionException(
                "Data source is None, cannot deserialize."
            )

        pack: DataPack = DataPack.deserialize(
            data_source,
            serialize_method=self.configs.serialize_method,
            zip_pack=self.configs.zip_pack,
        )

        if pack is None:
            raise ProcessExecutionException(
                f"Cannot recover pack from the following data source: \n"
                f"{data_source}"
            )

        yield pack

    @classmethod
    def default_configs(cls):
        r"""This defines a basic configuration structure for reader.

        Here:

          - zip_pack (bool): whether to zip the data pack. The default value is
            False.

          - indent (int): None for not indented, if larger than 0, the JSON
            files will be written in the with the provided indention. The
            default value is None.

          - serialize_method: The method used to serialize the data. Current
            available options are `jsonpickle` and `pickle`. Default is
            `jsonpickle`.

        Returns:
            The default configuration of this writer.
        """
        return {
            "zip_pack": False,
            "indent": None,
            "serialize_method": "jsonpickle",
        }


class RawDataDeserializeReader(BaseDeserializeReader):
    """
    This reader assumes the data passed in are raw DataPack strings.
    """

    def _collect(self, data_list: List[str]) -> Iterator[str]:  # type: ignore
        yield from data_list

    def _parse_pack(self, data_source: str) -> Iterator[DataPack]:
        yield DataPack.from_string(data_source)  # type: ignore


class RecursiveDirectoryDeserializeReader(BaseDeserializeReader):
    """
    This reader find all the files under the directory and read each one as
    a DataPack. If the `suffix` configuration is provided, then only files that
    end with this suffix will be read, otherwise all files will be read.
    Compressed data are supported through smart open.
    """

    def _collect(self, data_dir: str) -> Iterator[str]:  # type: ignore
        """
        This function will collect the files of the given directory. If the
         'suffix' field in the config is set, it will only take files matching
         that suffix. See :func:`~forte.data.readers.RecursiveDirectory
         DeserializeReader.default_configs` for the default configs.

        Args:
            data_dir: The root directory to search for the data packs.

        Returns:
            Iterator of the data pack string from the directory.
        """
        if not os.path.exists(data_dir):
            raise ValueError(
                f"The provided directory [{data_dir}] does not " f"exist."
            )
        for root, _, files in os.walk(data_dir):
            for file in files:
                if not self.configs.suffix or file.endswith(
                    self.configs.suffix
                ):
                    yield os.path.join(root, file)

    @classmethod
    def default_configs(cls):
        """
        Store the configs for this reader.

        .. code-block:: python

            {
                "suffix": ".json",
            }

        Here, "suffix" is used to finds files matching the suffix. The default
        value is `.json`. If None, then all files will be read.

        Returns:
            None
        """
        return {
            "suffix": ".json",
        }


class SinglePackReader(BaseDeserializeReader):
    """
    This reader reader one file of the given path as a DataPack. Compressed
    files are supported.
    """

    def _collect(self, data_path: str) -> Iterator[str]:  # type: ignore
        """
        This function will collect data path as a single file.

        Args:
            data_path: The file to the data pack.

        Returns:
            Iterator of only one item, the data pack string itself.
        """
        yield data_path


class MultiPackDeserializerBase(MultiPackReader):
    """
    This is a base implementation of deserializing multipacks, extend this
    reader by implementing the following functions:
      - Implement :func:`_get_multipack_str` to return the raw string of the
         multi pack from the data source (read from :func:`_collect`)
      - Implement :func:`_get_pack_str` to get the string of a particular
         data pack given the pack id.

    The data source (such as dataset path) should be passed in from the
     configs during `initialize`, since the sources for multipack is likely to
     be complex.

    The base reader will then construct the multipack based on these
      information.
    """

    def _collect(self, *args: Any, **kwargs: Any) -> Iterator[Any]:
        """
        This collect actually do not need any data source, it directly reads
        the data from the configurations.

        Returns:

        """
        for s in self._get_multipack_content(*args, **kwargs):
            yield s

    def _parse_pack(self, multi_pack_source: Any) -> Iterator[MultiPack]:
        m_pack: MultiPack = self._parse_multi_pack(multi_pack_source)

        for pid in m_pack.pack_ids():
            data_pack = self._get_pack(pid)
            if data_pack is None:
                logging.warning(
                    "Cannot locate the data pack with pid %d "
                    "for multi pack %d",
                    pid,
                    m_pack.pack_id,
                )
                break
            # Only in deserialization we can do this.
            m_pack.packs.append(data_pack)
        else:
            # No multi pack will be yield if there are packs not located.
            yield m_pack

    def _parse_multi_pack(self, multi_pack_source: Any) -> MultiPack:
        return MultiPack.deserialize(
            multi_pack_source,
            self.configs.serialize_method,
            self.configs.zip_pack,
        )

    @abstractmethod
    def _get_multipack_content(
        self, *args: Any, **kwargs: Any
    ) -> Iterator[Any]:
        """
        Implementation of this method should be responsible for yielding
        the multi packs.

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def _get_pack(self, pack_id: int) -> Optional[DataPack]:
        """
        Implementation of this method should be responsible for returning the
        actual data pack based on the pack id.

        Args:
            pack_id: representing the id of the data pack.

        Returns:
            The content of this data pack. You can either:
              - return the raw data pack string.
              - return the data pack as parsed DataPack object.

        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls):
        r"""This defines a basic configuration structure for writer.

        Here:
          - serialize_method: The method used to serialize the data. Current
              available options are `jsonpickle` and `pickle`. Default is
              `jsonpickle`.

        Returns: The default configuration of this writer.
        """
        return {
            "serialize_method": "jsonpickle",
        }


class MultiPackDirectoryReader(MultiPackDeserializerBase):
    """
    This reader implements one particular way of deserializing Multipack, which
    can be used to read the output written by
    :class:`~forte.processors.base.writers.PackNameMultiPackWriter`. It assumes
    the multipack are stored in a directory, and the data packs are stored in
    a directory too (they can be the same directory).
    """

    def __init__(self):
        super().__init__()
        self._open = open

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if self.configs.zip_pack:
            self._open = gzip.open

    def _get_multipack_content(self) -> Iterator[str]:  # type: ignore
        # pylint: disable=protected-access
        for mp_path in os.listdir(self.configs.multi_pack_dir):
            if mp_path.endswith(self.configs.suffix):
                yield os.path.join(self.configs.multi_pack_dir, mp_path)

    def _get_pack(self, pack_id: int) -> Optional[DataPack]:
        pack_path = os.path.join(
            self.configs.data_pack_dir, f"{pack_id}{self.configs.suffix}"
        )
        if os.path.exists(pack_path):
            return DataPack.deserialize(
                pack_path,
                serialize_method=self.configs.serialize_method,
                zip_pack=self.configs.zip_pack,
            )
        else:
            return None

    @classmethod
    def default_configs(cls):
        """
        Defines the default configuration for the multi pack reader.

        Here:
          - multi_pack_dir (str): the directory specifying the path storing the
              main multi pack content

          - data_pack_dir (str) : the directory specifying the path storing the
              data pack content.

          - suffix (str): the suffix of the data packs to be read.

          - serialize_method (str): The method used to serialize the data, this
              should be the same as how serialization is done. The current
              options are `jsonpickle` and `pickle`. The default method
              is `jsonpickle`.

          - zip_pack (bool): whether to zip the data pack. The default value is
              False.
        Returns:

        """
        return {
            "multi_pack_dir": None,
            "data_pack_dir": None,
            "suffix": ".json",
            "serialize_method": "jsonpickle",
            "zip_pack": False,
        }


# A short name for this class.
DirPackReader = RecursiveDirectoryDeserializeReader



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file implements AttributeExtractor, which is used to extract feature
from the attribute of entries.
"""
from typing import (
    Any,
    Union,
    Iterable,
    Hashable,
    SupportsInt,
    Dict,
    Optional,
    List,
)

from forte.common import ProcessorConfigError
from forte.common.configuration import Config
from forte.data.base_extractor import BaseExtractor
from forte.data.converter.feature import Feature
from forte.data.data_pack import DataPack
from forte.data.ontology.core import Entry
from forte.data.ontology.top import Annotation

__all__ = ["AttributeExtractor"]


class AttributeExtractor(BaseExtractor):
    r"""`AttributeExtractor` extracts feature from the attribute of entry.
    Most of the time, a user will not need to call this class explicitly,
    they will be called by the framework.
    """

    def initialize(self, config: Union[Dict, Config]):
        super().initialize(config)
        if self.config.entry_type is None:
            raise ProcessorConfigError(
                "The ``entry_type`` configuration must be "
                "provided and cannot be None."
            )
        if self.config.attribute is None:
            raise ProcessorConfigError(
                "The `attribute` configuration must be "
                "provided and cannot be None."
            )

    @classmethod
    def default_configs(cls):
        r"""Returns a dictionary of default hyper-parameters.


        Here:

        - "`attribute`": str
          The name of the attribute we want to extract from the entry. This
          attribute should present in the entry definition. There are some
          built-in attributes for some instance, such as `text` for
          `Annotation` entries. ``tid`` should be also available for any
          entries. The default value is ``tid``.
        - "`entry_type`": str
          The fully qualified name of the entry to extract attributes from.
          The default value is None, but this value must present or an
          `ProcessorConfigError` will be thrown.
        """
        config = super().default_configs()
        config.update(
            {
                "attribute": "tid",
                "entry_type": None,
            }
        )
        return config

    @classmethod
    def _get_attribute(cls, entry: Entry, attr: str) -> Any:
        r"""Get the attribute from entry. You can
        overwrite this function if you have special way to get the
        attribute from entry.

        Args:
            entry: An instance of Entry type, where the
                attribute will be extracted from.
            attr: The name of the attribute.

        Returns:
            Any. The attribute extracted from entry.
        """
        return getattr(entry, attr)

    @classmethod
    def _set_attribute(cls, entry: Entry, attr: str, value: Any):
        r"""Set the attribute of an entry to value.
        You can overwrite this function if you have special way to
        set the attribute.

        Args:
            entry: An instance of Entry type, where the
                attribute will be set.
            attr: The name of the attribute.
            value: The value to be set for the attribute.
        """
        if attr == "text":
            raise AttributeError("text attribute of entry cannot be changed.")
        setattr(entry, attr, value)

    def update_vocab(
        self, pack: DataPack, context: Optional[Annotation] = None
    ):
        r"""Get all attributes of one instance and add them into the vocabulary.

        Args:
            pack: The data pack input to extract vocabulary.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        """
        entry: Entry
        for entry in pack.get(self.config.entry_type, context):
            # The following pylint skip due to a bug:
            # https://github.com/PyCQA/pylint/issues/3507
            # Hashable is not recognized the type.
            # pylint: disable=isinstance-second-argument-not-valid-type
            element = self._get_attribute(entry, self.config.attribute)
            if not isinstance(element, Hashable):
                raise AttributeError(
                    "Only hashable element can be"
                    "added into the vocabulary. Consider setting"
                    "vocab_method to be raw and do not call update_vocab"
                    "if you only need the raw attribute value without"
                    "converting them into index."
                )
            self.add(element)

    def extract(
        self, pack: DataPack, context: Optional[Annotation] = None
    ) -> Feature:
        """Extract the attribute of an entry of the configured entry type.
        The entry type is passed in from via extractor config `entry_type`.

        Args:
            pack: The datapack that contains the current instance.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.

        Returns:
            Features (attributes) for instance with in the provided
            context, they will be converted to the representation based on
            the vocabulary configuration.
        """
        data = []

        instance: Annotation
        for instance in pack.get(self.config.entry_type, context):
            value = self._get_attribute(instance, self.config.attribute)
            rep = self.element2repr(value) if self.vocab else value
            data.append(rep)

        meta_data = {
            "need_pad": self.config.need_pad,
            "pad_value": self.get_pad_value(),
            "dim": 1,
            "dtype": int if self.vocab else Any,
        }
        return Feature(data=data, metadata=meta_data, vocab=self.vocab)

    def pre_evaluation_action(
        self, pack: DataPack, context: Optional[Annotation]
    ):
        r"""This function is performed on the pack before the evaluation
        stage, allowing one to perform some actions before the evaluation.
        By default, this function will remove all attributes defined in the
        config (set them to None). You can overwrite this function by yourself.

        Args:
            pack: The datapack that contains the current
                instance.
            context: The context is an Annotation entry where
                data are extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        """
        entry: Entry
        for entry in pack.get(self.config.entry_type, context):
            self._set_attribute(entry, self.config.attribute, None)

    def add_to_pack(
        self,
        pack: DataPack,
        predictions: Iterable[SupportsInt],
        context: Optional[Annotation] = None,
    ):
        r"""Add the prediction for attributes to the data pack.
        We assume the number of predictions in the iterable to be the same as
        the number of the entries of the defined type in the data pack.

        Args:
            pack: The datapack that contains the current
                instance.
            predictions: This is the output
                of the model, which should be the class index for the attribute.
            context: The context is an Annotation
                entry where predictions will be added to. This has the same
                meaning with `context` as in
                :meth:`~forte.data.base_extractor.BaseExtractor.extract`.
                If None, then the whole data pack will be used as the
                context. Default is None.
        """
        instance_entries: List[Entry] = list(
            pack.get(self.config.entry_type, context)
        )

        # The following pylint skip due to a bug:
        # https://github.com/PyCQA/pylint/issues/3507
        # Iterable is not recognized the type.
        # pylint: disable=isinstance-second-argument-not-valid-type
        # _predictions = predictions if isinstance(predictions, Iterable) else \
        #     [predictions]
        # if not isinstance(predictions, Iterable):
        #     predictions = [predictions]
        values = [self.id2element(int(x)) for x in predictions]
        for entry, value in zip(instance_entries, values):
            self._set_attribute(entry, self.config.attribute, value)



#  Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provide data across multiple data packs during training. A data pack iterator
iterates over each single data example across multiple data packs. A data pack
data set represents the dataset of a bunch of data packs. A raw example
represents a single data point in the dataset. A feature collection represents
an extracted feature corresponding to an input data point.
"""
from typing import Dict, Iterator, Type, Optional, List, Tuple, Union, Any

from asyml_utilities.hyperparams import HParams


from forte.data.converter import Converter
from forte.data.converter import Feature
from forte.data.data_pack import DataPack
from forte.data.base_extractor import BaseExtractor
from forte.data.ontology.core import EntryType
from forte.data.ontology.top import Annotation
from forte.data.types import DataRequest
from forte.utils import create_import_error_msg

try:
    import torch
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "extractor", "data pack dataset")
    ) from e

try:
    from texar.torch.data import IterDataSource, DatasetBase, Batch
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "extractor", "data pack dataset"
        )
    ) from e


__all__ = [
    "DataPackIterator",
    "DataPackDataset",
    "RawExample",
    "FeatureCollection",
]

# An instance is a single data point from data pack
RawExample = Tuple[int, DataPack]
FeatureCollection = Dict[str, Feature]


class DataPackIterator(IterDataSource):
    """
    An iterator generating data example from a stream of data packs.

    Args:
        pack_iterator: An iterator of
            :class:`~forte.data.data_pack.DataPack`.
        context_type: The granularity of a single example which
            could be any :class:`~forte.data.ontology.top.Annotation` type. For example, it can be
            :class:`~ft.onto.base_ontology.Sentence`, then each training example
            will represent the information of a sentence.
        request: The request of type `Dict` sent to
            :class:`~forte.data.data_pack.DataPack` to query
            specific data.
        skip_k: Will skip the first `skip_k` instances and generate
            data from the (`skip_k` + 1)th instance.

    Returns:
        An `Iterator` that each time produces a `Tuple` of an `tid`
        (of type `int`) and a data pack
        (of type :class:`~forte.data.data_pack.DataPack`).

    Here is an example usage:

        .. code-block:: python

            file_path: str = "data_samples/data_pack_dataset_test"
            reader = CoNLL03Reader()
            context_type = Sentence
            request = {Sentence: []}
            skip_k = 0

            train_pl: Pipeline = Pipeline()
            train_pl.set_reader(reader)
            train_pl.initialize()
            pack_iterator: Iterator[PackType] =
                train_pl.process_dataset(file_path)

            iterator: DataPackIterator = DataPackIterator(pack_iterator,
                                                          context_type,
                                                          request,
                                                          skip_k)

            for tid, data_pack in iterator:
                # process tid and data_pack

    .. note::
        For parameters `context_type`, `request`, `skip_k`, please refer to
        :meth:`~forte.data.data_pack.DataPack.get_data` in :class:`~forte.data.data_pack.DataPack`.
    """

    def __init__(
        self,
        pack_iterator: Iterator[DataPack],
        context_type: Type[Annotation],
        request: Optional[DataRequest] = None,
        skip_k: int = 0,
    ):
        super().__init__(self)

        self._get_data_args: Dict = {
            "context_type": context_type,
            "request": request,
            "skip_k": skip_k,
        }

        self._data_pack_iter: Iterator[DataPack] = pack_iterator
        self._instance_iter: Optional[Iterator[Dict[str, Any]]] = None
        self._curr_data_pack: Optional[DataPack] = None

    def __iter__(self):
        return self

    def __next__(self) -> RawExample:
        if self._curr_data_pack is None:
            self._curr_data_pack = next(self._data_pack_iter)
            self._instance_iter = self._curr_data_pack.get_data(
                **self._get_data_args
            )

        if self._instance_iter is None:
            raise ValueError("Instance iterator is None")

        try:
            return next(self._instance_iter)["tid"], self._curr_data_pack
        except StopIteration:
            # Current data pack has no more instance. Go to next data pack.
            self._curr_data_pack = next(self._data_pack_iter)
            self._instance_iter = self._curr_data_pack.get_data(
                **self._get_data_args
            )

        return next(self._instance_iter)["tid"], self._curr_data_pack


class DataPackDataset(DatasetBase):
    """
    A dataset representing data packs. Calling an
    :class:`~texar.torch.data.DataIterator`
    over this `DataPackDataset` will produce an `Iterate` over batch of examples
    parsed by a reader from given data packs.

    Args:
        data_source: A data source of type
            :class:`~forte.data.data_pack_dataset.DataPackIterator`.
        feature_schemes: A `Dict` containing all the information to do
            data pre-processing. This is exactly the same as the `schemes` in
            :meth:`~forte.train_preprocessor.TrainPreprocessor.request`.
        hparams: A `dict` or instance of
            :class:`~texar.torch.HParams` containing
            hyperparameters. See
            :meth:`~texar.torch.data.DatasetBase.default_hparams` in
            :class:`~texar.torch.data.DatasetBase` for the defaults.
        device: The device of the produced batches. For GPU training,
            set to current CUDA device.
    """

    def __init__(
        self,
        data_source: DataPackIterator,
        feature_schemes: Dict,
        hparams: Union[Dict, HParams] = None,
        device: Optional[torch.device] = None,
    ):
        self._data_source: DataPackIterator = data_source
        self._feature_scheme: Dict = feature_schemes

        super().__init__(self._data_source, hparams, device)

    def process(self, raw_example: RawExample) -> FeatureCollection:
        """
        Given an input which is a single data example, extract feature from it.

        Args:
            raw_example (tuple(dict, DataPack)): A `Tuple` where

              - The first element is a `Dict` produced by
                :meth:`~forte.data.data_pack.DataPack.get_data` in
                :class:`~forte.data.data_pack.DataPack`.

              - The second element is an instance of type
                :class:`~forte.data.data_pack.DataPack`.

        Returns:
            A `Dict` mapping from user-specified tags to the
            :class:`~forte.data.converter.Feature` extracted.

            .. note::
                Please refer to
                :meth:`~forte.train_preprocessor.TrainPreprocessor.request` for
                details about user-specified tags.
        """
        tid: int = raw_example[0]
        data_pack: DataPack = raw_example[1]
        instance_entry: EntryType = data_pack.get_entry(tid)  # type:ignore
        feature_collection: FeatureCollection = {}

        for tag, scheme in self._feature_scheme.items():
            extractor: BaseExtractor = scheme["extractor"]
            feature: Feature = extractor.extract(data_pack, instance_entry)
            feature_collection[tag] = feature

        return feature_collection

    def collate(self, examples: List[FeatureCollection]) -> Batch:
        """
        Given a batch of output from :meth:`process`, produce pre-processed
        data as well as masks and features.

        Args:
            examples: A `List` of result from :meth:`process`.

        Returns:
            A Texar Pytorch :class:`~Texar.torch.data.Batch`
            It can be treated as a `Dict` with the following structure:



            - `"data"`: List or `np.ndarray` or `torch.tensor`
              The pre-processed data.

              Please refer to :class:`~forte.data.converter.converter.Converter` for
              details.

            - `"masks"`: `np.ndarray` or `torch.tensor`
              All the masks for pre-processed data.

              Please refer to :class:`~forte.data.converter.converter.Converter` for
              details.

            - `"features"`: List[Feature]
              A List of :class:`~forte.data.converter.feature.Feature`. This
              is useful when users want to do customized pre-processing.

              Please refer to :class:`~forte.data.converter.Feature` for
              details.

            .. code-block:: python

                {
                    "tag_a": {
                        "data": <tensor>,
                        "masks": [<tensor1>, <tensor2>, ...],
                        "features": [<feature1>, <feature2>, ...]
                    },
                    "tag_b": {
                        "data": Tensor,
                        "masks": [<tensor1>, <tensor2>, ...],
                        "features": [<feature1>, <feature2>, ...]
                    }
                }


            .. note::
                The first level key in returned `batch` is the user-specified
                tags. Please refer to
                :meth:`~forte.train_preprocessor.TrainPreprocessor.request` for
                details about user-specified tags.
        """
        batch_size = len(examples)

        example_collection: Dict[str, List] = {}
        for example in examples:
            for tag, feature in example.items():
                if tag not in example_collection:
                    example_collection[tag] = []
                example_collection[tag].append(feature)

        tensor_collection: Dict[str, Dict[str, Any]] = {}
        for tag, features in example_collection.items():
            if tag not in tensor_collection:
                tensor_collection[tag] = {}

            converter: Converter = self._feature_scheme[tag]["converter"]
            data, masks = converter.convert(features)
            tensor_collection[tag]["data"] = data
            tensor_collection[tag]["masks"] = masks
            tensor_collection[tag]["features"] = features

        return Batch(batch_size, **tensor_collection)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
from typing import Any, Iterator, Dict, Set, Tuple

from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Document, MRCQuestion, Phrase
from ftx.onto.race_qa import Passage

__all__ = [
    "SquadReader",
]


class SquadReader(PackReader):
    r"""Reader for processing Stanford Question Answering Dataset (SQuAD).

    Stanford Question Answering Dataset (SQuAD) is a reading comprehension dataset,
    consisting of questions posed by crowdworkers on a set of Wikipedia articles,
    where the answer to every question is a segment of text, or span.

    Dataset can be downloaded at https://rajpurkar.github.io/SQuAD-explorer/.

    SquadReader reads each paragraph in the dataset as a separate Document, and the questions
    are concatenated behind the paragraph, form a Passage.
    Phrase are MRC answers marked as text spans. Each MRCQuestion has a list of answers.
    """

    def _collect(self, file_path: str) -> Iterator[Any]:  # type: ignore
        r"""Given file_path to the dataset, return an iterator to every data point in it.

        Args:
            file_path: path to the JSON file

        Returns: QA pairs and the context of a paragraph of a passage in SQuAD dataset.
        """
        with open(file_path, "r", encoding="utf8", errors="ignore") as file:
            jsonf = json.load(file)
            for dic in jsonf["data"]:
                title = dic["title"]
                cnt = 0
                for qa_dic in dic["paragraphs"]:
                    yield title + str(cnt), qa_dic["qas"], qa_dic["context"]
                    cnt += 1

    def _cache_key_function(self, text_file: str) -> str:
        return os.path.basename(text_file)

    def _parse_pack(self, qa_dict: Tuple[str, list, str]) -> Iterator[DataPack]:
        title, qas, context = qa_dict
        context_end = len(context)
        offset = context_end + 1
        text = context

        pack = DataPack()  # one datapack for a context
        for qa in qas:
            if qa["is_impossible"] is True:
                continue
            ques_text = qa["question"]
            ans = qa["answers"]
            text += "\n" + ques_text
            ques_end = offset + len(ques_text)
            question = MRCQuestion(pack, offset, ques_end)
            question.qid = qa["id"]
            offset = ques_end + 1
            for a in ans:
                ans_text = a["text"]
                ans_start = a["answer_start"]
                answer = Phrase(pack, ans_start, ans_start + len(ans_text))
                question.answers.append(answer)

        pack.set_text(text)

        passage = Passage(pack, 0, context_end)
        Document(pack, 0, len(pack.text))

        passage.passage_id = title
        pack.pack_name = title
        yield pack

    @classmethod
    def default_configs(cls):
        return {"file_ext": ".txt"}

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `PlainTextReader` which is
        `ft.onto.base_ontology.Document` with an empty set
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        record_meta["ft.onto.base_ontology.Document"] = set()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
    Module to automatically generate python ontology given json file
    Performs a preliminary check of dependencies
"""
import copy
import json
import logging
import os
import shutil
import tempfile
import warnings
from collections import defaultdict
from datetime import datetime
from distutils import dir_util
from pathlib import Path
from types import ModuleType
from typing import Dict, List, Optional, Tuple, Set, no_type_check, Any

import jsonschema
import typed_ast.ast3 as ast
import typed_astunparse as ast_unparse
from numpy import ndarray

from forte.data.ontology import top, utils
from forte.data.ontology.code_generation_exceptions import (
    DuplicateEntriesWarning,
    OntologySpecError,
    OntologyAlreadyGeneratedException,
    ParentEntryNotDeclaredException,
    TypeNotDeclaredException,
    UnsupportedTypeException,
    InvalidIdentifierException,
    DuplicatedAttributesWarning,
    ParentEntryNotSupportedException,
    OntologySpecValidationError,
    OntologySourceNotFoundException,
)
from forte.data.ontology.code_generation_objects import (
    NdArrayProperty,
    NonCompositeProperty,
    ListProperty,
    ClassTypeDefinition,
    EntryDefinition,
    Property,
    ImportManagerPool,
    EntryName,
    ModuleWriterPool,
    ImportManager,
    DictProperty,
    EntryTree,
)

# Builtin and local imports required in the generated python modules.
from forte.data.ontology.ontology_code_const import (
    REQUIRED_IMPORTS,
    DEFAULT_CONSTRAINTS_KEYS,
    AUTO_GEN_SIGNATURE,
    DEFAULT_PREFIX,
    SchemaKeywords,
    file_header,
    NON_COMPOSITES,
    COMPOSITES,
    ALL_INBUILT_TYPES,
    TOP_MOST_MODULE_NAME,
    PACK_TYPE_CLASS_NAME,
    hardcoded_pack_map,
    AUTO_GEN_FILENAME,
    AUTO_DEL_FILENAME,
    RESERVED_ATTRIBUTE_NAMES,
)
from forte.utils.utils_io import get_resource


def name_validation(name):
    parts = name.split(".")
    for part in parts:
        if not part.isidentifier():
            raise SyntaxError(f"'{part}' is not an valid identifier.")


def analyze_packages(packages: Set[str]):
    r"""Analyze the package paths to make sure they are valid.

    Args:
        packages: The list of packages.

    Returns: A list of the package paths, sorted by the package depth (
        deepest first).

    """
    package_len = []
    for p in packages:
        parts = p.split(".")
        package_len.append((len(parts), p))
        try:
            name_validation(p)
        except InvalidIdentifierException:
            logging.error(
                "Error analyzing package name: '%s' is "
                "not a valid package name",
                p,
            )
            raise

    return [p for (l, p) in sorted(package_len, reverse=True)]


def validate_entry(
    entry_name: str, sorted_packages: List[str], lenient_prefix=False
):
    """
    Validate if this entry name can be used. It currently checks for:
      1) If the package name is defined. (This can be turn off by setting
        `lenient_prefix` to True.
      2) If the entry name have at least 3 segments.

    Args:
        entry_name: The name to be validated.
        sorted_packages: The package names that are allowed.
        lenient_prefix: Whether we enforce that the entry must follow the
          pre-defined package name.

    Returns:

    """
    if not lenient_prefix:
        for package_name in sorted_packages:
            if entry_name.startswith(package_name):
                break
        else:
            # None of the package name matches.
            raise InvalidIdentifierException(
                f"Entry name [{entry_name}] does not start with any predefined "
                f"packages, please define the packages by using "
                f"`additional_prefixes` in the ontology. Or you can use the "
                f"default prefix 'ft.onto'."
            )

    entry_splits = entry_name.split(".")

    for e in entry_splits:
        if not e.isidentifier():
            raise InvalidIdentifierException(
                f"The entry name segment {e} is not a valid python identifier."
            )

    if len(entry_splits) < 3:
        raise InvalidIdentifierException(
            f"We currently require each entry to contains at least 3 levels, "
            f"which corresponds to the directory name, the file (module) name,"
            f"the entry class name. There are only {len(entry_splits)}"
            f"levels in [{entry_name}]."
        )


def as_init_str(init_args):
    """
    Create the __init__ string by using unparse in ast.
    Args:
        init_args: The ast args object of the init arguments.

    Returns:

    """
    # Unparsing the `__init__` args and normalising the string
    args = ast_unparse.unparse(init_args).strip().split(",", 1)
    return args[1].strip().replace("  ", "")


def is_composite_type(item_type: str):
    return item_type in COMPOSITES


def valid_composite_key(item_type: str):
    return item_type in ("int", "str")


class OntologyCodeGenerator:
    r"""Class to generate python ontology given ontology config in json format
    Salient Features -
        (1) Generates a class for each entry in the module corresponding to
        the defined entry package.
        (2) The entries of `forte.data.ontology.top`
        serve as ancestors of the user-defined entries.
        (2) Dependencies to other json configs through the json `imports`
    Example:

        .. code-block:: python

            destination_dir = OntologyCodeGenerator().generate(
                ... 'forte/ontology_specs/base_ontology.json')

    """

    def __init__(
        self, import_dirs: Optional[List[str]] = None, generate_all=False
    ):
        """
        Args:
            import_dirs: Additional user provided paths to search the
                imported json configs or installed ontologies from. By default
                paths provided in the imports, current working directory and
                the path where forte is installed (if it is) would be searched.
            generate_all: whether to generate all the packages even if some are
                already generated.
        """
        # The entries of the `self.top_ontology_module` serve as ancestors of
        # the user-defined entries.
        top_ontology_module: ModuleType = top

        # Builtin and local imports required in the generated python modules.
        self.required_imports: List[str] = REQUIRED_IMPORTS
        self.required_imports.append(top_ontology_module.__name__)

        # A collection of import managers: each manager is responsible for
        # controlling the imports of one module. The key of the collection is
        # the module's full name, e.g. ft.onto.base_ontology
        self.import_managers: ImportManagerPool = ImportManagerPool()

        # Store information to write each modules.
        self.module_writers: ModuleWriterPool = ModuleWriterPool(
            self.import_managers
        )

        # Mapping from entries parsed from the `top_ontology_module`
        # (default is `forte.data.ontology.top.py`), to their
        # `__init__` arguments.
        # self.top_init_args_strs: Dict[str, str] = {}
        self.root_base_entries: Set[str] = set()

        # Map from the full class name, to the list contains objects of
        # <typed_ast._ast3.arg>, which are the init arguments.
        self.top_init_args: Dict[str, Any] = {}

        # Mapping from user extendable entries to their ancestors
        self.top_to_core_entries: Dict[str, Set[str]] = {}

        # Mapping from user-defined entries to a set of base ancestor entries.
        # these base entries are the top level entries where user can
        # initialize, so they wil be part of the generated class's __init__
        self.base_entry_lookup: Dict[str, str] = {}

        # Populate the two dictionaries above. And make the classes in the base
        # ontology aware to the root manager.
        self.initialize_top_entries(
            self.import_managers.root, top_ontology_module
        )

        # A few pre-requesite type to support.
        self.import_managers.add_default_import("dataclasses.dataclass")
        self.import_managers.root.add_object_to_import("typing.Optional")

        for type_class in NON_COMPOSITES.values():
            self.import_managers.root.add_object_to_import(type_class)

        # Adjacency list to store the allowed types (in-built or user-defined),
        # and their attributes (if any) in order to validate the attribute
        # types.
        self.allowed_types_tree: Dict[str, Set] = {}
        for type_str in ALL_INBUILT_TYPES:
            self.allowed_types_tree[type_str] = set()

        self.installed_forte_dir = utils.get_installed_forte_dir()
        self.exclude_from_writing: Set[str] = set()

        # Directories to be examined to find json schema or installed ontology
        # packages which the user wants to import.
        self.import_dirs: List[str] = []

        # User defined directories are top choices.
        if import_dirs is not None:
            self.import_dirs.extend(import_dirs)

        # The current directory is secondary.
        self.import_dirs.append(os.getcwd())

        spec_base = "forte/ontology_specs"
        forte_spec_dir = get_resource(spec_base, False)

        # Lastly, the Forte installed directory.
        self.import_dirs.append(forte_spec_dir)
        self.exclude_from_writing = set()

        if not generate_all:
            logging.info(
                "Checking existing specification " "directory: %s",
                forte_spec_dir,
            )
            for existing_spec in os.listdir(forte_spec_dir):
                if existing_spec.endswith(".json"):
                    logging.info(
                        "Forte library contains %s, " "will skip this one.",
                        existing_spec,
                    )
                    self.exclude_from_writing.add(
                        os.path.join(spec_base, existing_spec)
                    )

    @no_type_check
    def initialize_top_entries(
        self, manager: ImportManager, base_ontology_module: ModuleType
    ):
        """
        Parses the file corresponding to `base_ontology_module` -
        (1) Imports the imports defined by the base file,
        (2) Imports the public API defined by the base file in it's `__all__`
        attribute. The imports are added to the import manager.

        (3) Extracts the name and inheritance of the class definitions and
        populates `self.top_to_core_entries`,
        (4) Extracts `__init__` arguments of class definitions and populates
        `self.top_init_args`
        (5) Includes type annotations for the `__init__` arguments.

        Args:
            manager: The import manager to be populated.
            base_ontology_module: File path of the module to be parsed.

        Returns:
        """
        tree = None
        with open(
            base_ontology_module.__file__, "r", encoding="utf-8"
        ) as base_ontology_file:
            tree = ast.parse(base_ontology_file.read())

        if tree is None:
            raise RuntimeError(
                f"Fail to load AST Tree from {base_ontology_module.__file__}"
            )

        base_module_name = base_ontology_module.__name__

        # Record a map from the import name to the full name.
        full_names = {}

        for elem in tree.body:
            # Adding all the imports.
            if isinstance(elem, ast.Import):
                for import_ in elem.names:
                    as_name = import_.asname
                    import_name = import_.name if as_name is None else as_name
                    full_names[import_name] = import_.name
                    manager.add_object_to_import(import_.name)

            if isinstance(elem, ast.ImportFrom):
                for import_ in elem.names:
                    full_names[import_.name] = f"{elem.module}.{import_.name}"
                    full_name = f"{elem.module}.{import_.name}"
                    manager.add_object_to_import(full_name)

            # Adding all the module objects defined in `__all__` to imports.
            if isinstance(elem, ast.Assign) and len(elem.targets) > 0:
                if elem.targets[0].id == "__all__":
                    full_names.update(
                        [
                            (
                                name.s,
                                f"{base_ontology_module.__name__}.{name.s}",
                            )
                            for name in elem.value.elts
                        ]
                    )

                    for name in elem.value.elts:
                        full_class_name = f"{base_module_name}.{name.s}"
                        manager.add_object_to_import(full_class_name)

            # Adding `__init__` arguments for each class
            if isinstance(elem, ast.ClassDef):
                # Adding base names for each class
                elem_base_names = set()
                if elem.bases is not None and len(elem.bases) > 0:
                    for elem_base in elem.bases:
                        while isinstance(elem_base, ast.Subscript):
                            # TODO: Doesn't handle typed class well.
                            elem_base_names.add(elem_base.slice.value.id)
                            elem_base_names.add(elem_base.value.id)
                            elem_base = elem_base.slice.value
                        elem_base_names.add(elem_base.id)
                init_func = None

                for func in elem.body:
                    if (
                        isinstance(func, ast.FunctionDef)
                        and func.name == "__init__"
                    ):
                        init_func = func
                        break

                if init_func is None:
                    warnings.warn(
                        f"No `__init__` function found in the class"
                        f" {elem.name} of the module "
                        f"{base_ontology_module}."
                    )
                else:
                    full_ele_name = full_names[elem.name]

                    # Assuming no variable args and keyword only args present in
                    # the base ontology module
                    for arg in init_func.args.args:
                        # Parsing the nested list of arg annotations
                        if arg.annotation is not None:
                            arg_ann = arg.annotation
                            while isinstance(arg_ann, ast.Subscript):
                                module = arg_ann.value.id
                                if module is not None and module in full_names:
                                    arg_ann.value.id = full_names[module]
                                arg_ann = arg_ann.slice.value
                            module = arg_ann.id

                            if module is not None and module in full_names:
                                arg_ann.id = full_names[module]
                                manager.add_object_to_import(arg_ann.id)

                                # Convert from PackType to more concrete pack
                                # type, such as DataPack or MultiPack.
                                if arg_ann.id == PACK_TYPE_CLASS_NAME:
                                    pack_class = hardcoded_pack_map(
                                        full_ele_name
                                    )
                                    manager.add_object_to_import(pack_class)
                                    arg_ann.id = pack_class

                    self.top_to_core_entries[full_ele_name] = elem_base_names
                    self.base_entry_lookup[full_ele_name] = full_ele_name
                    self.top_init_args[full_ele_name] = init_func.args

    def generate(
        self,
        spec_path: str,
        destination_dir: str = os.getcwd(),
        is_dry_run: bool = False,
        merged_path: Optional[str] = None,
        lenient_prefix=False,
        namespace_depth: int = -1,
    ) -> Optional[str]:
        r"""Function to generate and save the python ontology code after reading
        ontology from the input json file. This is the main entry point to
        the class.

        Args:
            spec_path: The input ontology specification file, which should
                be a json file.
            destination_dir: The folder in which config packages are to be
                generated. If not provided, current working directory is
                used. Ignored if `is_dry_run` is `True`.
            is_dry_run: if `True`, creates the ontology in the temporary
                directory, else, creates the ontology in the
                `destination_dir`.
            merged_path: if a path is provided, a merged ontology file will
                be written at this path.
            lenient_prefix: if `True`, will not enforce the entry name to
                match a known prefix.
            namespace_depth: set an integer argument namespace_depth to allow
              customized number of levels of namespace packaging.
              The generation of __init__.py for all the directory
              levels above namespace_depth will be disabled.
              For example, if we have an ontology level1.levle2.level3.
              something and namespace_depth=2, then we remove __init__.py
              under level1 and level1/level2 while keeping __init__.py under
              level1/level2/level3.
              When namespace_depth<=0, we just disable namespace packaging
              and include __init__.py in all directory levels.

        Returns:
            Directory path in which the modules are created: either one of
            the temporary directory or `destination_dir`.

        """
        # Update the list of directories to be examined for imported configs
        self.import_dirs.append(os.path.dirname(os.path.realpath(spec_path)))

        merged_schemas: List[Dict] = []
        merged_prefixes: List[str] = []

        # Generate ontology classes for the input json config and the configs
        # it is dependent upon.
        try:
            self.parse_ontology_spec(
                spec_path,
                merged_schema=merged_schemas,
                merged_prefixes=merged_prefixes,
                lenient_prefix=lenient_prefix,
            )
        except OntologySpecError:
            logging.error("Error at parsing [%s]", spec_path)
            raise

        # Now generate all data.

        # A temporary directory to save the generated file structure until the
        # generation is completed and verified.
        tempdir = tempfile.mkdtemp()
        # Starting from here, we won't add any more modules to import.
        self.import_managers.fix_all_modules()

        logging.info("Working on %s", spec_path)
        for writer in self.module_writers.writers():
            logging.info("Writing module: %s", writer.module_name)
            writer.write(tempdir, destination_dir, namespace_depth)
            logging.info("Done writing.")

        if merged_path is not None:
            logging.info("Writing merged schema at %s", merged_path)
            merged_config = {
                "name": "all_ontology",
                "definitions": merged_schemas,
                "additional_prefixes": list(set(merged_prefixes)),
            }
            with open(merged_path, "w", encoding="utf-8") as out:
                json.dump(merged_config, out, indent=2)
            logging.info("Done writing.")

        # When everything is successfully completed, copy the contents of
        # `self.tempdir` to the provided folder.
        if not is_dry_run:
            generated_top_dirs = set(utils.get_top_level_dirs(tempdir))
            for existing_top_dir in utils.get_top_level_dirs(destination_dir):
                if existing_top_dir in generated_top_dirs:
                    logging.warning(
                        "The directory with the name "
                        "%s is already present in "
                        "%s. New files will be merge into the "
                        "existing directory.",
                        existing_top_dir,
                        destination_dir,
                    )

            utils.copytree(
                tempdir,
                destination_dir,
                ignore_pattern_if_file_exists="*/__init__.py",
            )
            return destination_dir

        return tempdir

    def visit_ontology_imports(
        self,
        import_path: str,
        visited_paths: Optional[Dict[str, bool]] = None,
        rec_visited_paths: Optional[Dict[str, bool]] = None,
    ) -> Optional[Tuple[str, Dict[str, bool], Dict[str, bool]]]:
        # Initialize the visited dicts when the function is called for the
        # first time.
        if visited_paths is None:
            visited_paths = defaultdict(lambda: False)

        if rec_visited_paths is None:
            rec_visited_paths = defaultdict(lambda: False)

        # Check for import cycles
        if rec_visited_paths[import_path]:
            raise OntologyAlreadyGeneratedException(
                f"Ontology corresponding to {import_path} already "
                f"generated, cycles not permitted, aborting"
            )

        # If the ontology is already generated, need not generate it again
        if visited_paths[import_path]:
            return None

        # Add the json_file path to the visited dictionaries
        visited_paths[import_path] = True
        rec_visited_paths[import_path] = True

        # Validate and load the ontology specification.
        try:
            utils.validate_json_schema(import_path)
        except Exception as exception:
            if type(exception).__name__.split(".", maxsplit=1)[
                0
            ] == jsonschema.__name__ and hasattr(exception, "message"):
                raise OntologySpecValidationError() from exception
            raise

        return import_path, visited_paths, rec_visited_paths

    def find_import_path(self, import_path):
        for import_dir in self.import_dirs:
            full_spec_path = os.path.join(import_dir, import_path)
            if os.path.exists(full_spec_path):
                return full_spec_path

        raise OntologySourceNotFoundException(
            "Cannot find import [%s]." % import_path
        )

    def parse_ontology_spec(
        self,
        ontology_path: str,
        merged_schema: List[Dict],
        merged_prefixes: List[str],
        visited_paths: Optional[Dict[str, bool]] = None,
        rec_visited_paths: Optional[Dict[str, bool]] = None,
        lenient_prefix=False,
    ):
        r"""Performs a topological traversal on the directed graph formed by the
        imported json configs. While processing each config, it first generates
        the classes corresponding to the entries of the imported configs, then
        imports the generated python classes to generate the classes
        corresponding to the entries of `json_file_path`.
        Args:
            ontology_path: Path to the ontology.
            merged_schema: A list that store all the schema definitions.
            merged_prefixes: A list of prefixes from all schemas.
            visited_paths: Keeps track of the json configs already processed.
            rec_visited_paths: Keeps track of the current recursion stack, to
                detect, and throw error if any cycles are present.
            lenient_prefix: Whether to relax the requirement on the prefix.
        Returns:
        """
        import_info = self.visit_ontology_imports(
            ontology_path, visited_paths, rec_visited_paths
        )

        if import_info is None:
            return

        json_file_path, visited_paths, rec_visited_paths = import_info

        with open(json_file_path, "r", encoding="utf-8") as f:
            spec_dict = json.load(f)

        # Parse imported ontologies. Users can import them via a path relative
        # to the PYTHONPATH.
        relative_imports: Set[str] = set(
            spec_dict.get(SchemaKeywords.imports, [])
        )

        for rel_import in relative_imports:
            full_pkg_path: str = self.find_import_path(rel_import)
            logging.info("Imported ontology at: %s", full_pkg_path)
            self.parse_ontology_spec(
                full_pkg_path,
                merged_schema,
                merged_prefixes,
                visited_paths=visited_paths,
                rec_visited_paths=rec_visited_paths,
                lenient_prefix=lenient_prefix,
            )

        # Once the ontology for all the imported files is generated, generate
        # ontology of the current file.
        # Print relative json path in the ontology if the current directory is
        # the installation directory - example, when running the test cases
        curr_forte_dir = utils.get_current_forte_dir()

        print_json_file = json_file_path
        if self.installed_forte_dir is not None and os.path.samefile(
            curr_forte_dir, self.installed_forte_dir
        ):
            print_json_file = os.path.relpath(json_file_path, curr_forte_dir)

        self.parse_schema(
            spec_dict,
            print_json_file,
            merged_schema,
            merged_prefixes,
            lenient_prefix,
        )

        rec_visited_paths[json_file_path] = False

    def parse_schema_for_no_import_onto_specs_file(
        self,
        ontology_path: str,
        ontology_dict: Dict,
        lenient_prefix=False,
        merged_entry_tree: Optional[EntryTree] = None,
    ):
        r"""Function to populate the `merged_entry_tree` after reading
            ontology from the input json file.

        Args:
            ontology_path: The path to the input ontology specification file,
                which should be a json file, and it should have all the entries
                inside with no import as key.
            ontology_dict: The loaded dictionary of ontology specifications
            lenient_prefix: Whether to remove the constraint on the prefix set.
            merged_entry_tree: an EntryTree type object and if it's not`None`
                then after running this function, all the entries from
                ontology specification file would be parsed into a tree
                structure with parent and children entries to represent
                the relationship.

        Returns:

        """
        # Assume no import key in the ontology file
        if SchemaKeywords.imports in ontology_dict.keys():
            raise OntologySpecValidationError(
                "The system cannot build entry trees from an ontology file "
                "that imports other ontologies. But we find `import` keywords "
                f"in the provided ontology {ontology_path}"
                " Please provide a merged ontology file with all imports "
                "resolved by using the command "
                "`generate_ontology create --merged_path`"
            )

        merged_schema: List[Dict] = []
        merged_prefixes: List[str] = []

        self.parse_schema(
            ontology_dict,
            ontology_path,
            merged_schema,
            merged_prefixes,
            lenient_prefix,
            merged_entry_tree,
        )

    def parse_schema(
        self,
        schema: Dict,
        source_json_file: str,
        merged_schema: List[Dict],
        merged_prefixes: List[str],
        lenient_prefix=False,
        merged_entry_tree: Optional[EntryTree] = None,
    ):
        r"""Generates ontology code for a parsed schema extracted from a
        json config. Appends entry code to the corresponding module. Creates a
        new module file if module is generated for the first time.

        Args:
            schema: Ontology dictionary extracted from a json config.
            source_json_file: Path of the source json file.
            merged_schema: The merged schema is used to remember all
                definitions during parsing.
            merged_prefixes: To remember all prefixes encountered during
                parsing.
            lenient_prefix: Whether to remove the constraint on the prefix set.
            merged_entry_tree: an EntryTree type object and if it's not`None`
                then after running this function, all the entries from
                ontology specification file would be parsed into a tree
                structure with parent and children entries to represent
                the relationship.
        Returns:
            Modules to be imported by dependencies of the current ontology.
        """
        entry_definitions: List[Dict] = schema[SchemaKeywords.definitions]
        merged_schema.extend(entry_definitions)

        if SchemaKeywords.prefixes in schema:
            merged_prefixes.extend(schema[SchemaKeywords.prefixes])

        allowed_packages = set(
            schema.get(SchemaKeywords.prefixes, []) + [DEFAULT_PREFIX]
        )
        sorted_prefixes = analyze_packages(allowed_packages)

        file_desc = file_header(
            schema.get(SchemaKeywords.description, ""),
            schema.get(SchemaKeywords.ontology_name, ""),
        )

        for definition in entry_definitions:
            raw_entry_name = definition[SchemaKeywords.entry_name]
            validate_entry(raw_entry_name, sorted_prefixes, lenient_prefix)

            if raw_entry_name in self.allowed_types_tree:
                warnings.warn(
                    f"Class {raw_entry_name} already present in the "
                    f"ontology, will be overridden.",
                    DuplicateEntriesWarning,
                )
            self.allowed_types_tree[raw_entry_name] = set()

            # Add the entry definition to the import managers.
            # This time adding to the root manager so everyone can access it
            # if needed, but they will only appear in the import list when
            # requested.
            # Entry class should be added to the imports before the attributes
            # to be able to used as the attribute type for the same entry.
            self.import_managers.root.add_object_to_import(raw_entry_name)

            # Get various parts of the entry name.
            en = EntryName(raw_entry_name)
            entry_item, properties = self.parse_entry(en, definition)

            # Add it as a defining object.
            self.import_managers.get(en.module_name).add_defining_objects(
                raw_entry_name
            )

            # Get or set module writer only if the ontology to be generated
            # is not already installed.
            if source_json_file not in self.exclude_from_writing:
                module_writer = self.module_writers.get(en.module_name)
                module_writer.set_description(file_desc)
                module_writer.source_file = source_json_file
                # Add entry item to the writer.
                module_writer.add_entry(en, entry_item)

            # Adding entry attributes to the allowed types for validation.
            for property_name in properties:
                # Check if the name is allowed.
                if not property_name.isidentifier():
                    raise InvalidIdentifierException(
                        f"The property name: {property_name} is not a valid "
                        f"python identifier."
                    )

                if property_name in self.allowed_types_tree[en.class_name]:
                    warnings.warn(
                        f"Attribute type for the entry {en.class_name} "
                        f"and the attribute {property_name} already present in "
                        f"the ontology, will be overridden",
                        DuplicatedAttributesWarning,
                    )
                self.allowed_types_tree[en.class_name].add(property_name)
            # populate the entry tree based on information
            if merged_entry_tree is not None:
                curr_entry_name = en.class_name
                parent_entry_name = definition["parent_entry"]
                curr_entry_attributes = self.allowed_types_tree[en.class_name]
                merged_entry_tree.add_node(
                    curr_entry_name, parent_entry_name, curr_entry_attributes
                )

    def cleanup_generated_ontology(
        self, path, is_forced=False
    ) -> (Tuple[bool, Optional[str]]):
        """
        Deletes the generated files and directories. Generated files are
        identified by the header `***automatically_generated***`. Generated
        directories are identified by `.generated` empty marker files. Moves
        the files to a timestamped folder inside `.deleted` folder located in
        the parent directory path by default.
        Args:
            path: Path of the directory to be searched for generated files.
            is_forced: Deletes the generated files and directories without
             moving them to `.deleted` directory. `False` by default.
        Returns:
            Whether `path` is empty after the cleanup operation is
        completed
            The timestamped `.deleted` directory path
        """
        path = os.path.abspath(path)

        rel_paths = dir_util.copy_tree(path, "", dry_run=1)
        rel_paths = [
            os.path.dirname(file)
            for file in rel_paths
            if os.path.basename(file).startswith(AUTO_GEN_FILENAME)
        ]

        del_dir = None
        if not is_forced:
            curr_time_str = datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S-%f")
            del_dir = os.path.join(
                os.path.dirname(path), AUTO_DEL_FILENAME, curr_time_str
            )
            for rel_path in rel_paths:
                joined_path = os.path.join(del_dir, rel_path)
                Path(joined_path).mkdir(parents=True, exist_ok=True)
        rel_paths += [""]
        return (
            self._cleanup_generated_ontology(path, "", del_dir, rel_paths),
            del_dir,
        )

    def _cleanup_generated_ontology(
        self, outer_path, relative_path, delete_dir, allowed_relative_paths
    ) -> bool:
        """
        Recursively deletes the generated files and the newly empty directories.
        Args:
            outer_path: Path of the directory to be searched for generated files
            relative_path: Path relative to outer path.
            delete_dir: Directory where the delete file structure has to be
            moved.
            allowed_relative_paths: Directory paths having `.generated` marker
            files.

        Returns: Whether `path` is empty after the cleanup operation is
        completed.
        """
        if os.path.dirname(relative_path) not in allowed_relative_paths:
            return False

        path = os.path.join(outer_path, relative_path)
        dst_dir = (
            os.path.join(delete_dir, os.path.dirname(relative_path))
            if delete_dir is not None
            else None
        )

        if os.path.isfile(path):
            # path is a file type
            # delete .generated marker files and automatically generated files
            is_empty = os.path.basename(path).startswith(AUTO_GEN_FILENAME)
            if not is_empty and os.access(path, os.R_OK):
                with open(path, "r", encoding="utf-8") as f:
                    lines = f.readlines()
                    if len(lines) > 0:
                        if lines[0].startswith(f"# {AUTO_GEN_SIGNATURE}"):
                            is_empty = True
            if is_empty:
                if delete_dir is not None:
                    shutil.copy(path, dst_dir)
                os.unlink(path)
        else:
            # path is a directory type
            is_empty = True
            for child in os.listdir(path):
                child_rel_path = os.path.join(relative_path, child)
                if not self._cleanup_generated_ontology(
                    outer_path,
                    child_rel_path,
                    delete_dir,
                    allowed_relative_paths,
                ):
                    is_empty = False
            if is_empty:
                if delete_dir is not None:
                    dir_util.copy_tree(path, dst_dir)
                os.rmdir(path)
        return is_empty

    def replace_annotation(self, entry_name: EntryName, func_args):
        this_manager = self.import_managers.get(entry_name.module_name)
        for arg in func_args.args:
            if arg.annotation is not None:
                arg_ann = arg.annotation
                # Handling the type name for cases like Optional[X]
                while isinstance(arg_ann, ast.Subscript):
                    # The types for arg_ann and so on are in typed_ast._ast3,
                    # these types are protected hence hard to be used here.

                    short_ann_name = arg_ann.value.id  # type: ignore
                    full_ann_name: str = this_manager.get_name_to_use(
                        short_ann_name
                    )

                    arg_ann.value.id = full_ann_name  # type: ignore
                    arg_ann = arg_ann.slice.value  # type: ignore

                # Handling the type name for arguments.
                arg_ann.id = this_manager.get_name_to_use(arg_ann.id)

    def construct_init(self, entry_name: EntryName, base_entry: str):
        base_init_args = self.top_init_args[base_entry]
        custom_init_args = copy.deepcopy(base_init_args)

        self.replace_annotation(entry_name, custom_init_args)
        custom_init_args_str = as_init_str(custom_init_args)
        return custom_init_args_str

    def parse_entry(
        self, entry_name: EntryName, schema: Dict
    ) -> Tuple[EntryDefinition, List[str]]:
        """
        Args:
            entry_name: Object holds various name form of the entry.
            schema: Dictionary containing specifications for an entry.

        Returns: extracted entry information: entry package string, entry
        filename, entry class entry_name, generated entry code and entry
        attribute names.
        """
        this_manager = self.import_managers.get(entry_name.module_name)

        # Determine the parent entry of this entry.
        parent_entry: str = schema[SchemaKeywords.parent_entry]

        if parent_entry.startswith(TOP_MOST_MODULE_NAME):
            raise ParentEntryNotSupportedException(
                f"The parent entry {parent_entry} cannot be directly inherited,"
                f" please inherit a type from {top.__name__} or your own"
                f" ontology."
            )

        if not this_manager.is_imported(parent_entry):
            raise ParentEntryNotDeclaredException(
                f"The parent entry {parent_entry} is not declared. It is "
                f"neither in the base entries nor in custom entries. "
                f"Please check them ontology specification, and make sure the "
                f"entry is defined before this."
            )

        base_entry: Optional[str] = self.find_base_entry(
            entry_name.class_name, parent_entry
        )

        if base_entry is None:
            raise OntologySpecError(
                f"Cannot find the base entry for entry "
                f"{entry_name.class_name} and {parent_entry}"
            )

        if base_entry not in self.top_init_args:
            raise ParentEntryNotSupportedException(
                f"Cannot add {entry_name.class_name} to the ontology as "
                f"it's parent entry {parent_entry} is not supported. This is "
                f"likely that the entries are not inheriting the allowed types."
            )

        # Take the property definitions of this entry.
        properties: List[Dict] = schema.get(SchemaKeywords.attributes, [])

        this_manager = self.import_managers.get(entry_name.module_name)

        # Validate if the parent entry is present.
        if not this_manager.is_known_name(parent_entry):
            raise ParentEntryNotDeclaredException(
                f"Cannot add {entry_name.class_name} to the ontology as "
                f"it's parent entry {parent_entry} is not present "
                f"in the ontology."
            )

        parent_entry_use_name = this_manager.get_name_to_use(parent_entry)

        property_items, property_names = [], []
        for prop_schema in properties:
            # TODO: add test
            prop_name = prop_schema["name"]
            if prop_name in RESERVED_ATTRIBUTE_NAMES:
                raise InvalidIdentifierException(
                    f"The attribute name {prop_name} is reserved and cannot be "
                    f"used, please consider changed the name. The list of "
                    f"reserved name strings are "
                    f"{RESERVED_ATTRIBUTE_NAMES}"
                )

            property_names.append(prop_schema["name"])
            property_items.append(self.parse_property(entry_name, prop_schema))

        # For special classes that requires a constraint.
        core_bases: Set[str] = self.top_to_core_entries[base_entry]
        entry_constraint_keys: Dict[str, str] = {}
        if any(item == "BaseLink" for item in core_bases):
            entry_constraint_keys = DEFAULT_CONSTRAINTS_KEYS["BaseLink"]
        elif any(item == "BaseGroup" for item in core_bases):
            entry_constraint_keys = DEFAULT_CONSTRAINTS_KEYS["BaseGroup"]

        class_att_items: List[ClassTypeDefinition] = []
        for schema_key, class_key in entry_constraint_keys.items():
            if schema_key in schema:
                constraint_type_ = schema[schema_key]
                constraint_type_name = this_manager.get_name_to_use(
                    constraint_type_
                )

                if constraint_type_name is None:
                    raise TypeNotDeclaredException(
                        f"The type {constraint_type_} is not defined but it is "
                        f"specified in {schema_key} of the definition of "
                        f"{schema['entry_name']}. Please define them before "
                        f"this entry type."
                    )

                # TODO: cannot handle constraints that contain self-references.
                # self_ref = entry_name.class_name == constraint_type_

                class_att_items.append(
                    ClassTypeDefinition(class_key, constraint_type_name)
                )

        # TODO: Can assign better object type to Link and Group objects
        custom_init_arg_str: str = self.construct_init(entry_name, base_entry)

        entry_item = EntryDefinition(
            name=entry_name.name,
            class_type=parent_entry_use_name,
            init_args=custom_init_arg_str,
            properties=property_items,
            class_attributes=class_att_items,
            description=schema.get(SchemaKeywords.description, None),
        )

        return entry_item, property_names

    def parse_ndarray(
        self,
        manager: ImportManager,
        schema: Dict,
        att_name: str,
        desc: str,
    ):
        ndarray_dtype = None
        if SchemaKeywords.ndarray_dtype in schema:
            ndarray_dtype = schema[SchemaKeywords.ndarray_dtype]

        ndarray_shape = None
        if SchemaKeywords.ndarray_shape in schema:
            ndarray_shape = schema[SchemaKeywords.ndarray_shape]

        if ndarray_dtype is None or ndarray_shape is None:
            warnings.warn(
                "Either dtype or shape is not specified."
                " It is recommended to specify both of them."
            )

        default_val = None
        if ndarray_dtype and ndarray_shape:
            default_val = ndarray(ndarray_shape, dtype=ndarray_dtype)

        return NdArrayProperty(
            manager,
            att_name,
            ndarray_dtype,
            ndarray_shape,
            description=desc,
            default_val=default_val,
        )

    def parse_dict(
        self,
        manager: ImportManager,
        schema: Dict,
        entry_name: EntryName,
        att_name: str,
        att_type: str,
        desc: str,
    ):
        if (
            SchemaKeywords.dict_key_type not in schema
            or SchemaKeywords.dict_value_type not in schema
        ):
            raise TypeNotDeclaredException(
                f"Item type of the attribute {att_name} for the entry "
                f" {entry_name.class_name} not declared. This attribute is "
                f"a composite type: {att_type}, it should have a "
                f"{SchemaKeywords.dict_key_type} and "
                f"{SchemaKeywords.dict_value_type}."
            )

        key_type = schema[SchemaKeywords.dict_key_type]
        if not valid_composite_key(key_type):
            raise UnsupportedTypeException(
                f"Key type {key_type} for entry {entry_name.name}'s "
                f"attribute {att_name} is not supported, we only support a "
                f"limited set of keys."
            )

        value_type = schema[SchemaKeywords.dict_value_type]
        if is_composite_type(value_type):
            # Case of nested.
            raise UnsupportedTypeException(
                f"Item type {value_type} for entry {entry_name.name}'s "
                f"attribute {att_name} is a composite type, we do not support "
                f"nested composite type."
            )

        if not manager.is_known_name(value_type):
            # Case of unknown.
            raise TypeNotDeclaredException(
                f"Item type {value_type} for the entry "
                f"{entry_name.name} of the attribute {att_name} "
                f"not declared in ontology."
            )

        # Make sure the import of these related types are handled.
        manager.add_object_to_import(value_type)

        self_ref = entry_name.class_name == value_type

        default_val: Dict = {}

        return DictProperty(
            manager,
            att_name,
            key_type,
            value_type,
            description=desc,
            default_val=default_val,
            self_ref=self_ref,
        )

    def parse_list(
        self,
        manager: ImportManager,
        schema: Dict,
        entry_name: EntryName,
        att_name: str,
        att_type: str,
        desc: str,
    ) -> ListProperty:
        if SchemaKeywords.element_type not in schema:
            raise TypeNotDeclaredException(
                f"Item type for the attribute {att_name} of the entry "
                f"[{entry_name.class_name}] not declared. This attribute is "
                f"a composite type: {att_type}, it should have a "
                f"{SchemaKeywords.element_type}."
            )

        item_type = schema[SchemaKeywords.element_type]
        if is_composite_type(item_type):
            # Case of nested.
            raise UnsupportedTypeException(
                f"Item type {item_type} for entry {entry_name.name}'s "
                f"attribute {att_name} is a composite type, we do not support "
                f"nested composite type."
            )

        if not manager.is_known_name(item_type):
            # Case of unknown.
            raise TypeNotDeclaredException(
                f"Item type {item_type} for the entry "
                f"{entry_name.name} of the attribute {att_name} "
                f"not declared in ontology."
            )

        # Make sure the import of these related types are handled.
        manager.add_object_to_import(item_type)

        self_ref = entry_name.class_name == item_type

        return ListProperty(
            manager,
            att_name,
            item_type,
            description=desc,
            default_val=[],
            self_ref=self_ref,
        )

    def parse_non_composite(
        self,
        manager: ImportManager,
        att_name: str,
        att_type: str,
        desc: str,
        default_val: str,
        self_ref: bool = False,
    ) -> NonCompositeProperty:
        manager.add_object_to_import("typing.Optional")

        return NonCompositeProperty(
            manager,
            att_name,
            att_type,
            description=desc,
            default_val=default_val,
            self_ref=self_ref,
        )

    def parse_property(self, entry_name: EntryName, schema: Dict) -> Property:
        """
        Parses instance and class properties defined in an entry schema and
        checks for the constraints allowed by the ontology generation system.

        Args:
            entry_name: Entry Name object that contains various form of the
              entry's name.
            schema: Entry definition schema
        Returns: An object of class `code_generation_util.FileItem` containing
         the generated code.
        """
        att_name = schema[SchemaKeywords.attribute_name]
        att_type = schema[SchemaKeywords.attribute_type]

        manager: ImportManager = self.import_managers.get(
            entry_name.module_name
        )

        # schema type should be present in the validation tree
        # TODO: Remove this hack
        if not manager.is_known_name(att_type):
            raise TypeNotDeclaredException(
                f"Attribute type '{att_type}' for the entry "
                f"'{entry_name.name}' of the attribute '{att_name}' not "
                f"declared in the ontology"
            )

        desc = schema.get(SchemaKeywords.description, None)
        default_val = schema.get(SchemaKeywords.default_value, None)

        # element type should be present in the validation tree
        if att_type in COMPOSITES:
            if att_type == "List":
                return self.parse_list(
                    manager, schema, entry_name, att_name, att_type, desc
                )
            elif att_type == "Dict":
                return self.parse_dict(
                    manager, schema, entry_name, att_name, att_type, desc
                )
            elif att_type == "NdArray":
                return self.parse_ndarray(manager, schema, att_name, desc)
        elif att_type in NON_COMPOSITES or manager.is_imported(att_type):
            self_ref = entry_name.class_name == att_type
            return self.parse_non_composite(
                manager,
                att_name,
                att_type,
                desc,
                default_val,
                self_ref=self_ref,
            )

        raise UnsupportedTypeException(f"{att_type} is not a supported type.")

    def find_base_entry(
        self, this_entry: str, parent_entry: str
    ) -> Optional[str]:
        """Find the `base_entry`. As a side effect, it will populate the
        internal state `self.base_entry_lookup`. The base will be one of the
        predefined base entry group in the top ontology, which are:
         ["Link", "Annotation", "Group", "Generic", "MultiPackLink",
         "MultiPackGroup", "MultiPackGeneric"]

        Args:
            this_entry: the entry name for which the base_entry is to be
               returned.
            parent_entry: parent of the `this_entry`. Note that the `base_entry`
             of the `this_entry` is same as the base entry of the
             `parent_entry`.

        Returns:
            `base_entry` for the entry `this_entry`.

        Example:
            If the subclass structure is -
            `Word` inherits `Token` inherits `Annotation`.
            The base_entry for both `Word` and `Token` should be `Annotation`.
        """
        base_entry: Optional[str]
        if parent_entry in self.top_init_args:
            # The top init args contains the objects in the top.py.
            base_entry = parent_entry
        else:
            base_entry = self.base_entry_lookup.get(parent_entry, None)

        if base_entry is not None:
            self.base_entry_lookup[this_entry] = base_entry

        return base_entry



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    Union,
    Any,
    Set,
    Callable,
    Tuple,
)

import numpy as np
from sortedcontainers import SortedList

from forte.common.exception import (
    ProcessExecutionException,
    UnknownOntologyClassException,
)
from forte.data import data_utils_io
from forte.data.base_pack import BaseMeta, BasePack
from forte.data.index import BaseIndex
from forte.data.ontology.core import Entry
from forte.data.ontology.core import EntryType
from forte.data.ontology.top import (
    Annotation,
    Link,
    Group,
    SinglePackEntries,
    Generics,
    AudioAnnotation,
)
from forte.data.span import Span
from forte.data.types import ReplaceOperationsType, DataRequest
from forte.utils import get_class

logger = logging.getLogger(__name__)

__all__ = ["Meta", "DataPack", "DataIndex"]


class Meta(BaseMeta):
    r"""Basic Meta information associated with each instance of
    :class:`~forte.data.data_pack.DataPack`.

    Args:
        pack_name:  An name to identify the data pack, which is helpful in
           situation like serialization. It is suggested that the packs should
           have different doc ids.
        language: The language used by this data pack, default is English.
        span_unit: The unit used for interpreting the Span object of this
          data pack. Default is character.
        sample_rate: An integer specifying the sample rate of audio payload.
          Default is None.
        info: Store additional string based information that the user add.
    Attributes:
        pack_name:  storing the provided `pack_name`.
        language: storing the provided `language`.
        sample_rate: storing the provided `sample_rate`.
        info: storing the provided `info`.
        record: Initialized as a dictionary. This is not a required field.
            The key of the record should be the entry type and values should
            be attributes of the entry type. All the information would be used
            for consistency checking purpose if the pipeline is initialized with
            `enforce_consistency=True`.
    """

    def __init__(
        self,
        pack_name: Optional[str] = None,
        language: str = "eng",
        span_unit: str = "character",
        sample_rate: Optional[int] = None,
        info: Optional[Dict[str, str]] = None,
    ):
        super().__init__(pack_name)
        self.language = language
        self.span_unit = span_unit
        self.sample_rate: Optional[int] = sample_rate
        self.record: Dict[str, Set[str]] = {}
        self.info: Dict[str, str]
        if info is None:
            self.info = {}
        else:
            self.info = info


def as_entry_type(entry_type: Union[str, Type[EntryType]]):
    entry_type_: Type[EntryType]
    if isinstance(entry_type, str):
        entry_type_ = get_class(entry_type)
        if not issubclass(entry_type_, Entry):
            raise ValueError(
                f"The specified entry type [{entry_type}] "
                f"does not correspond to a "
                f"`forte.data.ontology.core.Entry` class"
            )
    else:
        entry_type_ = entry_type
    return entry_type_


def as_sorted_error_check(entries: List[EntryType]) -> SortedList:
    """
    Given a list of entries, return a sorted list of it. If unknown entry
    classes are seen during this process,
    a :class:`~forte.common.exception.UnknownOntologyClassException` exception will be
    thrown.

    Args:
        entries: A list of entries to be converted.

    Returns: Sorted list of the input entries.
    """
    try:
        return SortedList(entries)
    except TypeError as e:
        for entry in entries:
            if isinstance(entry, Dict) and "py/object" in entry:
                entry_class = entry["py/object"]
                try:
                    get_class(entry_class)
                except ValueError:
                    raise UnknownOntologyClassException(
                        f"Cannot deserialize ontology type {entry_class}, "
                        f"make sure it is included in the PYTHONPATH."
                    ) from e


class DataPack(BasePack[Entry, Link, Group]):
    # pylint: disable=too-many-public-methods, unused-private-member
    r"""A :class:`~forte.data.data_pack.DataPack` contains a piece of natural language text and a
    collection of NLP entries (annotations, links, and groups). The natural
    language text could be a document, paragraph or in any other granularity.

    Args:
        pack_name: A name for this data pack.
    """

    def __init__(self, pack_name: Optional[str] = None):
        super().__init__(pack_name)
        self._text = ""
        self._audio: Optional[np.ndarray] = None

        self.annotations: SortedList[Annotation] = SortedList()
        self.links: SortedList[Link] = SortedList()
        self.groups: SortedList[Group] = SortedList()
        self.generics: SortedList[Generics] = SortedList()
        self.audio_annotations: SortedList[AudioAnnotation] = SortedList()

        self.__replace_back_operations: ReplaceOperationsType = []
        self.__processed_original_spans: List[Tuple[Span, Span]] = []

        self.__orig_text_len: int = 0

        self._index: DataIndex = DataIndex()

    def __getstate__(self):
        r"""
        In serialization,
            1) will serialize the annotation sorted list as a normal list;
            2) will not serialize the indices
        """
        state = super().__getstate__()
        state["annotations"] = list(state["annotations"])
        state["links"] = list(state["links"])
        state["groups"] = list(state["groups"])
        state["generics"] = list(state["generics"])
        state["audio_annotations"] = list(state["audio_annotations"])
        return state

    def __setstate__(self, state):
        r"""
        In deserialization, we
            1) transform the annotation list back to a sorted list;
            2) initialize the indexes.
            3) Obtain the pack ids.
        """
        super().__setstate__(state)

        # For backward compatibility.
        if "replace_back_operations" in self.__dict__:
            self.__replace_back_operations = self.__dict__.pop(
                "replace_back_operations"
            )
        if "processed_original_spans" in self.__dict__:
            self.__processed_original_spans = self.__dict__.pop(
                "processed_original_spans"
            )
        if "orig_text_len" in self.__dict__:
            self.__orig_text_len = self.__dict__.pop("orig_text_len")

        self.annotations = as_sorted_error_check(self.annotations)
        self.links = as_sorted_error_check(self.links)
        self.groups = as_sorted_error_check(self.groups)
        self.generics = as_sorted_error_check(self.generics)

        # Add `hasattr` checking here for backward compatibility
        self.audio_annotations = (
            as_sorted_error_check(self.audio_annotations)
            if hasattr(self, "audio_annotations")
            else SortedList()
        )

        self._index = DataIndex()
        self._index.update_basic_index(list(self.annotations))
        self._index.update_basic_index(list(self.links))
        self._index.update_basic_index(list(self.groups))
        self._index.update_basic_index(list(self.generics))
        self._index.update_basic_index(list(self.audio_annotations))

        for a in self.annotations:
            a.set_pack(self)

        for a in self.links:
            a.set_pack(self)

        for a in self.groups:
            a.set_pack(self)

        for a in self.generics:
            a.set_pack(self)

        for a in self.audio_annotations:
            a.set_pack(self)

    def __iter__(self):
        yield from self.annotations
        yield from self.links
        yield from self.groups
        yield from self.generics
        yield from self.audio_annotations

    def _init_meta(self, pack_name: Optional[str] = None) -> Meta:
        return Meta(pack_name)

    def _validate(self, entry: EntryType) -> bool:
        return isinstance(entry, SinglePackEntries)

    @property
    def text(self) -> str:
        r"""Return the text of the data pack"""
        return self._text

    @property
    def audio(self) -> Optional[np.ndarray]:
        r"""Return the audio of the data pack"""
        return self._audio

    @property
    def sample_rate(self) -> Optional[int]:
        r"""Return the sample rate of the audio data"""
        return getattr(self._meta, "sample_rate")

    @property
    def all_annotations(self) -> Iterator[Annotation]:
        """
        An iterator of all annotations in this data pack.

        Returns: Iterator of all annotations, of
        type :class:`~forte.data.ontology.top.Annotation`.

        """
        yield from self.annotations

    @property
    def num_annotations(self) -> int:
        """
        Number of annotations in this data pack.

        Returns: (int) Number of the links.

        """
        return len(self.annotations)

    @property
    def all_links(self) -> Iterator[Link]:
        """
        An iterator of all links in this data pack.

        Returns: Iterator of all links, of
        type :class:`~forte.data.ontology.top.Link`.

        """
        yield from self.links

    @property
    def num_links(self) -> int:
        """
        Number of links in this data pack.

        Returns: Number of the links.

        """
        return len(self.links)

    @property
    def all_groups(self) -> Iterator[Group]:
        """
        An iterator of all groups in this data pack.

        Returns: Iterator of all groups, of
        type :class:`~forte.data.ontology.top.Group`.

        """
        yield from self.groups

    @property
    def num_groups(self):
        """
        Number of groups in this data pack.

        Returns: Number of groups.

        """
        return len(self.groups)

    @property
    def all_generic_entries(self) -> Iterator[Generics]:
        """
        An iterator of all generic entries in this data pack.

        Returns: Iterator of generic

        """
        yield from self.generics

    @property
    def num_generics_entries(self):
        """
        Number of generics entries in this data pack.

        Returns: Number of generics entries.

        """
        return len(self.generics)

    @property
    def all_audio_annotations(self) -> Iterator[AudioAnnotation]:
        """
        An iterator of all audio annotations in this data pack.

        Returns: Iterator of all audio annotations, of
        type :class:`~forte.data.ontology.top.AudioAnnotation`.

        """
        yield from self.audio_annotations

    @property
    def num_audio_annotations(self):
        """
        Number of audio annotations in this data pack.

        Returns: Number of audio annotations.

        """
        return len(self.audio_annotations)

    def get_span_text(self, begin: int, end: int) -> str:
        r"""Get the text in the data pack contained in the span.

        Args:
            begin: begin index to query.
            end: end index to query.

        Returns:
            The text within this span.
        """
        return self._text[begin:end]

    def get_span_audio(self, begin: int, end: int) -> np.ndarray:
        r"""Get the audio in the data pack contained in the span.
        `begin` and `end` represent the starting and ending indices of the span
        in audio payload respectively. Each index corresponds to one sample in
        audio time series.

        Args:
            begin: begin index to query.
            end: end index to query.

        Returns:
            The audio within this span.
        """
        if self._audio is None:
            raise ProcessExecutionException(
                "The audio payload of this DataPack is not set. Please call"
                " method `set_audio` before running `get_span_audio`."
            )
        return self._audio[begin:end]

    def set_text(
        self,
        text: str,
        replace_func: Optional[Callable[[str], ReplaceOperationsType]] = None,
    ):

        if len(text) < len(self._text):
            raise ProcessExecutionException(
                "The new text is overwriting the original one with shorter "
                "length, which might cause unexpected behavior."
            )

        if len(self._text):
            logging.warning(
                "Need to be cautious when changing the text of a "
                "data pack, existing entries may get affected. "
            )

        span_ops = [] if replace_func is None else replace_func(text)

        # The spans should be mutually exclusive
        (
            self._text,
            self.__replace_back_operations,
            self.__processed_original_spans,
            self.__orig_text_len,
        ) = data_utils_io.modify_text_and_track_ops(text, span_ops)

    def set_audio(self, audio: np.ndarray, sample_rate: int):
        r"""Set the audio payload and sample rate of the :class:`~forte.data.data_pack.DataPack`
        object.

        Args:
            audio: A numpy array storing the audio waveform.
            sample_rate: An integer specifying the sample rate.
        """
        self._audio = audio
        self.set_meta(sample_rate=sample_rate)

    def get_original_text(self):
        r"""Get original unmodified text from the :class:`~forte.data.data_pack.DataPack` object.

        Returns:
            Original text after applying the `replace_back_operations` of
            :class:`~forte.data.data_pack.DataPack` object to the modified text
        """
        original_text, _, _, _ = data_utils_io.modify_text_and_track_ops(
            self._text, self.__replace_back_operations
        )
        return original_text

    def get_original_span(
        self, input_processed_span: Span, align_mode: str = "relaxed"
    ):
        r"""Function to obtain span of the original text that aligns with the
        given span of the processed text.

        Args:

            input_processed_span: Span of the processed text for which
                the corresponding span of the original text is desired.
            align_mode: The strictness criteria for alignment in the
                ambiguous cases, that is, if a part of input_processed_span
                spans a part of the inserted span, then align_mode controls
                whether to use the span fully or ignore it completely according
                to the following possible values:

                    - "strict" - do not allow ambiguous input, give ValueError.
                    - "relaxed" - consider spans on both sides.
                    - "forward" - align looking forward, that is, ignore the
                      span towards the left, but consider the span towards
                      the right.
                    - "backward" - align looking backwards, that is, ignore the
                      span towards the right, but consider the span towards the
                      left.


        Returns:
            Span of the original text that aligns with input_processed_span

        Example:
            * Let o-up1, o-up2, ... and m-up1, m-up2, ... denote the unprocessed
              spans of the original and modified string respectively. Note that
              each o-up would have a corresponding m-up of the same size.
            * Let o-pr1, o-pr2, ... and m-pr1, m-pr2, ... denote the processed
              spans of the original and modified string respectively. Note that
              each o-p is modified to a corresponding m-pr that may be of a
              different size than o-pr.
            * Original string:
              <--o-up1--> <-o-pr1-> <----o-up2----> <----o-pr2----> <-o-up3->
            * Modified string:
              <--m-up1--> <----m-pr1----> <----m-up2----> <-m-pr2-> <-m-up3->
            * Note that `self.inverse_original_spans` that contains modified
              processed spans and their corresponding original spans, would look
              like - [(o-pr1, m-pr1), (o-pr2, m-pr2)]

        .. code-block:: python

            >> data_pack = DataPack()
            >> original_text = "He plays in the park"
            >> data_pack.set_text(original_text,\
            >>                    lambda _: [(Span(0, 2), "She"))]
            >> data_pack.text
            "She plays in the park"
            >> input_processed_span = Span(0, len("She plays"))
            >> orig_span = data_pack.get_original_span(input_processed_span)
            >> data_pack.get_original_text()[orig_span.begin: orig_span.end]
            "He plays"

        """
        assert align_mode in ["relaxed", "strict", "backward", "forward"]

        req_begin = input_processed_span.begin
        req_end = input_processed_span.end

        def get_original_index(
            input_index: int, is_begin_index: bool, mode: str
        ) -> int:
            r"""
            Args:
                input_index: begin or end index of the input span
                is_begin_index: if the index is the begin index of the input
                span or the end index of the input span
                mode: alignment mode
            Returns:
                Original index that aligns with input_index
            """
            if len(self.__processed_original_spans) == 0:
                return input_index

            len_processed_text = len(self._text)
            orig_index = None
            prev_end = 0
            for (
                inverse_span,
                original_span,
            ) in self.__processed_original_spans:
                # check if the input_index lies between one of the unprocessed
                # spans
                if prev_end <= input_index < inverse_span.begin:
                    increment = original_span.begin - inverse_span.begin
                    orig_index = input_index + increment
                # check if the input_index lies between one of the processed
                # spans
                elif inverse_span.begin <= input_index < inverse_span.end:
                    # look backward - backward shift of input_index
                    if is_begin_index and mode in ["backward", "relaxed"]:
                        orig_index = original_span.begin
                    if not is_begin_index and mode == "backward":
                        orig_index = original_span.begin - 1

                    # look forward - forward shift of input_index
                    if is_begin_index and mode == "forward":
                        orig_index = original_span.end
                    if not is_begin_index and mode in ["forward", "relaxed"]:
                        orig_index = original_span.end - 1

                # break if the original index is populated
                if orig_index is not None:
                    break
                prev_end = inverse_span.end

            if orig_index is None:
                # check if the input_index lies between the last unprocessed
                # span
                inverse_span, original_span = self.__processed_original_spans[
                    -1
                ]
                if inverse_span.end <= input_index < len_processed_text:
                    increment = original_span.end - inverse_span.end
                    orig_index = input_index + increment
                else:
                    # check if there input_index is not valid given the
                    # alignment mode or lies outside the processed string
                    raise ValueError(
                        f"The input span either does not adhere "
                        f"to the {align_mode} alignment mode or "
                        f"lies outside to the processed string."
                    )
            return orig_index

        orig_begin = get_original_index(req_begin, True, align_mode)
        orig_end = get_original_index(req_end - 1, False, align_mode) + 1

        return Span(orig_begin, orig_end)

    @classmethod
    def deserialize(
        cls,
        data_source: Union[Path, str],
        serialize_method: str = "jsonpickle",
        zip_pack: bool = False,
    ) -> "DataPack":
        """
        Deserialize a Data Pack from a string. This internally calls the
        internal :meth:`~forte.data.base_pack.BasePack._deserialize()` function
        from :class:`~forte.data.base_pack.BasePack`.

        Args:
            data_source: The path storing data source.
            serialize_method: The method used to serialize the data, this
              should be the same as how serialization is done. The current
              options are `jsonpickle` and `pickle`. The default method
              is `jsonpickle`.
            zip_pack: Boolean value indicating whether the input source is
              zipped.

        Returns:
            An data pack object deserialized from the string.
        """
        return cls._deserialize(data_source, serialize_method, zip_pack)

    def _add_entry(self, entry: EntryType) -> EntryType:
        r"""Force add an :class:`~forte.data.ontology.core.Entry` object to the
        :class:`~forte.data.data_pack.DataPack` object. Allow duplicate entries in a pack.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry`
                object to be added to the pack.

        Returns:
            The input entry itself
        """
        return self.__add_entry_with_check(entry, True)

    def __add_entry_with_check(
        self, entry: EntryType, allow_duplicate: bool = True
    ) -> EntryType:
        r"""Internal method to add an :class:`~forte.data.ontology.core.Entry`
        object to the :class:`~forte.data.DataPack` object.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry` object
                to be added to the datapack.
            allow_duplicate: Whether we allow duplicate in the datapack.

        Returns:
            The input entry itself
        """
        if isinstance(entry, Annotation):
            target = self.annotations

            begin, end = entry.begin, entry.end

            if begin < 0:
                raise ValueError(
                    f"The begin {begin} is smaller than 0, this"
                    f"is not a valid begin."
                )

            if end > len(self.text):
                if len(self.text) == 0:
                    raise ValueError(
                        f"The end {end} of span is greater than the text "
                        f"length {len(self.text)}, which is invalid. The text "
                        f"length is 0, so it may be the case the you haven't "
                        f"set text for the data pack. Please set the text "
                        f"before calling `add_entry` on the annotations."
                    )
                else:
                    pack_ref = entry.pack.pack_id
                    raise ValueError(
                        f"The end {end} of span is greater than the text "
                        f"length {len(self.text)}, which is invalid. The "
                        f"problematic entry is of type {entry.__class__} "
                        f"at [{begin}:{end}], in pack {pack_ref}."
                    )

        elif isinstance(entry, Link):
            target = self.links
        elif isinstance(entry, Group):
            target = self.groups
        elif isinstance(entry, Generics):
            target = self.generics
        elif isinstance(entry, AudioAnnotation):
            target = self.audio_annotations
        else:
            raise ValueError(
                f"Invalid entry type {type(entry)}. A valid entry "
                f"should be an instance of Annotation, Link, Group, Generics "
                "or AudioAnnotation."
            )

        if not allow_duplicate:
            index = target.index(entry)
            if index < 0:
                # Return the existing entry if duplicate is not allowed.
                return target[index]

        target.add(entry)
        # update the data pack index if needed
        self._index.update_basic_index([entry])
        if self._index.link_index_on and isinstance(entry, Link):
            self._index.update_link_index([entry])
        if self._index.group_index_on and isinstance(entry, Group):
            self._index.update_group_index([entry])
        self._index.deactivate_coverage_index()
        self._pending_entries.pop(entry.tid)
        return entry

    def delete_entry(self, entry: EntryType):
        r"""Delete an :class:`~forte.data.ontology.core.Entry` object from the
        :class:`~forte.data.data_pack.DataPack`. This find out the entry in the index and remove it
        from the index. Note that entries will only appear in the index if
        `add_entry` (or _add_entry_with_check) is called.

        Please note that deleting a entry do not guarantee the deletion of
        the related entries.

        Args:
            entry: An :class:`~forte.data.ontology.core.Entry`
                object to be deleted from the pack.

        """
        if isinstance(entry, Annotation):
            target = self.annotations
        elif isinstance(entry, Link):
            target = self.links
        elif isinstance(entry, Group):
            target = self.groups
        elif isinstance(entry, Generics):
            target = self.generics
        elif isinstance(entry, AudioAnnotation):
            target = self.audio_annotations
        else:
            raise ValueError(
                f"Invalid entry type {type(entry)}. A valid entry "
                f"should be an instance of Annotation, Link, or Group."
            )

        begin: int = target.bisect_left(entry)

        index_to_remove = -1
        for i, e in enumerate(target[begin:]):
            if e.tid == entry.tid:
                index_to_remove = begin + i
                break

        if index_to_remove < 0:
            logger.warning(
                "The entry with id %d that you are trying to removed "
                "does not exists in the data pack's index. Probably it is "
                "created but not added in the first place.",
                entry.tid,
            )
        else:
            target.pop(index_to_remove)

        # update basic index
        self._index.remove_entry(entry)

        # set other index invalid
        self._index.turn_link_index_switch(on=False)
        self._index.turn_group_index_switch(on=False)
        self._index.deactivate_coverage_index()

    @classmethod
    def validate_link(cls, entry: EntryType) -> bool:
        return isinstance(entry, Link)

    @classmethod
    def validate_group(cls, entry: EntryType) -> bool:
        return isinstance(entry, Group)

    def get_data(
        self,
        context_type: Union[str, Type[Annotation], Type[AudioAnnotation]],
        request: Optional[DataRequest] = None,
        skip_k: int = 0,
    ) -> Iterator[Dict[str, Any]]:
        r"""Fetch data from entries in the data_pack of type
        `context_type`. Data includes `"span"`, annotation-specific
        default data fields and specific data fields by `"request"`.

        Annotation-specific data fields means:

            - `"text"` for ``Type[Annotation]``
            - `"audio"` for ``Type[AudioAnnotation]``

        Currently, we do not support Groups and Generics in the request.

        Example:

            .. code-block:: python

                requests = {
                    base_ontology.Sentence:
                        {
                            "component": ["dummy"],
                            "fields": ["speaker"],
                        },
                    base_ontology.Token: ["pos", "sense"],
                    base_ontology.EntityMention: {
                    },
                }
                pack.get_data(base_ontology.Sentence, requests)

        Args:
            context_type:
                The granularity of the data context, which
                could be any :class:`~forte.data.ontology.top.Annotation` or
                :class:`~forte.data.ontology.top.AudioAnnotation` type.
                Behaviors under different context_type varies:

                - str type will be converted into either
                  :class:`~forte.data.ontology.top.Annotation` type or
                  :class:`~forte.data.ontology.top.AudioAnnotation` type.
                - ``Type[Annotation]``: the default data field for getting
                  context data is :attr:`text`. This function iterates
                  :attr:`all_annotations` to search target entry data.
                - ``Type[AudioAnnotation]``: the default data field for getting
                  context data is :attr:`audio` which stores audio data in
                  numpy arrays. This function iterates
                  :attr:`all_audio_annotations` to search target entry data.

            request: The
                entry types and fields User wants to request.
                The keys of the requests dict are the required entry types
                and the value should be either:

                - a list of field names or
                - a dict which accepts three keys: `"fields"`, `"component"`,
                  and `"unit"`.

                    - By setting `"fields"` (list), users
                      specify the requested fields of the entry. If "fields"
                      is not specified, only the default fields will be
                      returned.
                    - By setting `"component"` (list), users
                      can specify the components by which the entries are
                      generated. If `"component"` is not specified, will return
                      entries generated by all components.
                    - By setting `"unit"` (string), users can
                      specify a unit by which the annotations are indexed.

                Note that for all annotation types, `"span"`
                fields and annotation-specific data fields are returned by
                default.

                For all link types, `"child"` and `"parent"` fields are
                returned by default.
            skip_k: Will skip the first `skip_k` instances and generate
                data from the (`offset` + 1)th instance.

        Returns:
            A data generator, which generates one piece of data (a dict
            containing the required entries, fields, and context).
        """
        context_type_: Union[Type[Annotation], Type[AudioAnnotation]]
        if isinstance(context_type, str):
            context_type_ = get_class(context_type)
            if not issubclass(context_type_, Entry):
                raise ValueError(
                    f"The provided `context_type` [{context_type_}] "
                    f"is not a subclass to the"
                    f"`forte.data.ontology.top.Annotation` class"
                )
        else:
            context_type_ = context_type

        annotation_types: Dict[
            Union[Type[Annotation], Type[AudioAnnotation]], Union[Dict, List]
        ] = {}
        link_types: Dict[Type[Link], Union[Dict, List]] = {}
        group_types: Dict[Type[Group], Union[Dict, List]] = {}
        generics_types: Dict[Type[Generics], Union[Dict, List]] = {}
        audio_annotation_types: Dict[
            Type[AudioAnnotation], Union[Dict, List]
        ] = {}

        if request is not None:
            for key_, value in request.items():
                key = as_entry_type(key_)
                if issubclass(key, Annotation):
                    annotation_types[key] = value
                elif issubclass(key, Link):
                    link_types[key] = value
                elif issubclass(key, Group):
                    group_types[key] = value
                elif issubclass(key, Generics):
                    generics_types[key] = value
                elif issubclass(key, AudioAnnotation):
                    audio_annotation_types[key] = value

        context_args = annotation_types.get(context_type_)

        context_components, _, context_fields = self._parse_request_args(
            context_type_, context_args
        )

        valid_context_ids: Set[int] = self._index.query_by_type_subtype(
            context_type_
        )

        if context_components:
            valid_component_id: Set[int] = set()
            for component in context_components:
                valid_component_id |= self.get_ids_by_creator(component)
            valid_context_ids &= valid_component_id

        def get_annotation_list(
            c_type: Union[Type[Annotation], Type[AudioAnnotation]]
        ):
            r"""Get an annotation list of a given context type.

            Args:
                c_type:
                    The granularity of the data context, which
                    could be any :class:`~forte.data.ontology.top.Annotation` type.

            Raises:
                NotImplementedError: raised when the given context type is
                    not implemented.

            Returns:
                List(Union[Annotation, AudioAnnotation]):
                    a list of annotations which is a copy of `self.annotations`
                    and it enables modifications of `self.annotations` while
                    iterating through its copy.
            """
            if issubclass(c_type, Annotation):
                return list(self.annotations)
            elif issubclass(c_type, AudioAnnotation):
                return list(self.audio_annotations)
            else:
                raise NotImplementedError(
                    f"Context type is set to {c_type},"
                    " but currently we only support"
                    " [Annotation, AudioAnnotation]."
                )

        def get_context_data(c_type, context):
            r"""Get context-specific data of a given context type and
                context.

            Args:
                c_type:
                    The granularity of the data context, which
                    could be any :class:`~forte.data.ontology.top.Annotation` type.
                context: context that
                    contains data to be extracted.

            Raises:
                NotImplementedError: raised when the given context type is
                    not implemented.

            Returns:
                str: context data.
            """
            if issubclass(c_type, Annotation):
                return self.text[context.begin : context.end]
            elif issubclass(c_type, AudioAnnotation):
                return self.audio[context.begin : context.end]
            else:
                raise NotImplementedError(
                    f"Context type is set to {context_type}"
                    "but currently we only support"
                    "[Annotation, AudioAnnotation]"
                )

        skipped = 0
        for context in get_annotation_list(context_type_):
            if context.tid not in valid_context_ids or not isinstance(
                context, context_type_
            ):
                continue
            if skipped < skip_k:
                skipped += 1
                continue
            data: Dict[str, Any] = {}
            data["context"] = get_context_data(context_type_, context)
            data["offset"] = context.begin

            for field in context_fields:
                data[field] = getattr(context, field)

            if annotation_types:
                for a_type, a_args in annotation_types.items():
                    if issubclass(a_type, context_type_):
                        continue
                    if a_type.__name__ in data:
                        raise KeyError(
                            f"Requesting two types of entries with the "
                            f"same class name {a_type.__name__} at the "
                            f"same time is not allowed"
                        )
                    data[
                        a_type.__name__
                    ] = self._generate_annotation_entry_data(
                        a_type, a_args, data, context
                    )

            if audio_annotation_types:
                for a_type, a_args in audio_annotation_types.items():
                    if a_type.__name__ in data:
                        raise KeyError(
                            f"Requesting two types of entries with the "
                            f"same class name {a_type.__name__} at the "
                            f"same time is not allowed"
                        )
                    data[
                        a_type.__name__
                    ] = self._generate_annotation_entry_data(
                        a_type, a_args, data, context
                    )

            if link_types:
                for l_type, l_args in link_types.items():
                    if l_type.__name__ in data:
                        raise KeyError(
                            f"Requesting two types of entries with the "
                            f"same class name {l_type.__name__} at the "
                            f"same time is not allowed"
                        )
                    data[l_type.__name__] = self._generate_link_entry_data(
                        l_type, l_args, data, context
                    )
            # TODO: Getting Group based on range is not done yet.
            if group_types:
                raise NotImplementedError(
                    "Querying groups based on ranges is "
                    "currently not supported."
                )
            if generics_types:
                raise NotImplementedError(
                    "Querying generic types based on ranges is "
                    "currently not supported."
                )
            yield data

    def _parse_request_args(self, a_type, a_args):
        # request which fields generated by which component
        components = None
        unit = None
        fields = set()
        if isinstance(a_args, dict):
            components = a_args.get("component")
            # pylint: disable=isinstance-second-argument-not-valid-type
            # TODO: until fix: https://github.com/PyCQA/pylint/issues/3507
            if components is not None and not isinstance(components, Iterable):
                raise TypeError(
                    "Invalid request format for 'components'. "
                    "The value of 'components' should be of an iterable type."
                )
            unit = a_args.get("unit")
            if unit is not None and not isinstance(unit, str):
                raise TypeError(
                    "Invalid request format for 'unit'. "
                    "The value of 'unit' should be a string."
                )
            a_args = a_args.get("fields", set())

        # pylint: disable=isinstance-second-argument-not-valid-type
        # TODO: disable until fix: https://github.com/PyCQA/pylint/issues/3507
        if isinstance(a_args, Iterable):
            fields = set(a_args)
        elif a_args is not None:
            raise TypeError(
                f"Invalid request format for '{a_type}'. "
                f"The request should be of an iterable type or a dict."
            )

        fields.add("tid")
        return components, unit, fields

    def _generate_annotation_entry_data(
        self,
        a_type: Union[Type[Annotation], Type[AudioAnnotation]],
        a_args: Union[Dict, Iterable],
        data: Dict,
        cont: Optional[Annotation],
    ) -> Dict:

        components, unit, fields = self._parse_request_args(a_type, a_args)

        a_dict: Dict[str, Any] = {}
        a_dict["span"] = []
        # For AudioAnnotation, since the data is single numpy array
        # we don't initialize an empty list for a_dict["audio"]
        if issubclass(a_type, Annotation):
            a_dict["text"] = []
        elif issubclass(a_type, AudioAnnotation):
            a_dict["audio"] = []

        for field in fields:
            a_dict[field] = []
        unit_begin = 0
        if unit is not None:
            if unit not in data:
                raise KeyError(
                    f"{unit} is missing in data. You need to "
                    f"request {unit} before {a_type}."
                )
            a_dict["unit_span"] = []

        cont_begin = cont.begin if cont else 0
        annotation: Union[Type[Annotation], Type[AudioAnnotation]]
        for annotation in self.get(a_type, cont, components):  # type: ignore
            # we provide span, text (and also tid) by default
            a_dict["span"].append((annotation.begin, annotation.end))

            if isinstance(annotation, Annotation):
                a_dict["text"].append(annotation.text)
            elif isinstance(annotation, AudioAnnotation):
                a_dict["audio"].append(annotation.audio)
            else:
                raise NotImplementedError(
                    f"Annotation is set to {annotation}"
                    "but currently we only support"
                    "instances of [Annotation, "
                    "AudioAnnotation] and their subclass."
                )
            for field in fields:
                if field in ("span", "text", "audio"):
                    continue
                if field == "context_span":
                    a_dict[field].append(
                        (
                            annotation.begin - cont_begin,
                            annotation.end - cont_begin,
                        )
                    )
                    continue

                a_dict[field].append(getattr(annotation, field))

            if unit is not None:
                while not self._index.in_span(
                    data[unit]["tid"][unit_begin],
                    annotation.span,
                ):
                    unit_begin += 1

                unit_span_begin = unit_begin
                unit_span_end = unit_span_begin + 1

                while self._index.in_span(
                    data[unit]["tid"][unit_span_end],
                    annotation.span,
                ):
                    unit_span_end += 1

                a_dict["unit_span"].append((unit_span_begin, unit_span_end))
        for key, value in a_dict.items():
            a_dict[key] = np.array(value)

        return a_dict

    def _generate_link_entry_data(
        self,
        a_type: Type[Link],
        a_args: Union[Dict, Iterable],
        data: Dict,
        cont: Optional[Annotation],
    ) -> Dict:

        components, unit, fields = self._parse_request_args(a_type, a_args)

        if unit is not None:
            raise ValueError(f"Link entries cannot be indexed by {unit}.")

        a_dict: Dict[str, Any] = {}
        for field in fields:
            a_dict[field] = []
        a_dict["parent"] = []
        a_dict["child"] = []

        link: Link
        for link in self.get(a_type, cont, components):
            parent_type = link.ParentType.__name__
            child_type = link.ChildType.__name__

            if parent_type not in data:
                raise KeyError(
                    f"The Parent entry of {a_type} is not requested."
                    f" You should also request {parent_type} with "
                    f"{a_type}"
                )
            if child_type not in data:
                raise KeyError(
                    f"The child entry of {a_type} is not requested."
                    f" You should also request {child_type} with "
                    f"{a_type}"
                )

            a_dict["parent"].append(
                np.where(data[parent_type]["tid"] == link.parent)[0][0]
            )
            a_dict["child"].append(
                np.where(data[child_type]["tid"] == link.child)[0][0]
            )

            for field in fields:
                if field in ("parent", "child"):
                    continue

                a_dict[field].append(getattr(link, field))

        for key, value in a_dict.items():
            a_dict[key] = np.array(value)
        return a_dict

    def build_coverage_for(
        self,
        context_type: Type[Union[Annotation, AudioAnnotation]],
        covered_type: Type[EntryType],
    ):
        """
        User can call this function to build coverage index for specific types.
        The index provide a in-memory mapping from entries of `context_type`
        to the entries "covered" by it.
        See :class:`forte.data.data_pack.DataIndex` for more details.

        Args:
            context_type: The context/covering type.
            covered_type: The entry to find under the context type.

        """
        if self._index.coverage_index(context_type, covered_type) is None:
            self._index.build_coverage_index(self, context_type, covered_type)

    def covers(
        self,
        context_entry: Union[Annotation, AudioAnnotation],
        covered_entry: EntryType,
    ) -> bool:
        """
        Check if the `covered_entry` is covered (in span) of the `context_type`.

        See :meth:`~forte.data.data_pack.DataIndex.in_span` and
        :meth:`~forte.data.data_pack.DataIndex.in_audio_span` for the definition
        of `in span`.

        Args:
            context_entry: The context entry.
            covered_entry: The entry to be checked on whether it is in span
              of the context entry.

        Returns (bool): True if in span.
        """
        return covered_entry.tid in self._index.get_covered(
            self, context_entry, covered_entry.__class__
        )

    def iter_in_range(
        self,
        entry_type: Type[EntryType],
        range_annotation: Union[Annotation, AudioAnnotation],
    ) -> Iterator[EntryType]:
        """
        Iterate the entries of the provided type within or fulfill the
        constraints of the `range_annotation`. The constraint is True if
        an entry is :meth:`~forte.data.data_pack.DataIndex.in_span` or
        :meth:`~forte.data.data_pack.DataIndex.in_audio_span` of the provided
        `range_annotation`.

        Internally, if the coverage index between the entry type and the
        type of the `range_annotation` is built, then this will create the
        iterator from the index. Otherwise, the function will iterate them
        from scratch (which is slower). If there are frequent usage of this
        function, it is suggested to build the coverage index.

        Only when `range_annotation` is an instance of `AudioAnnotation` will
        the searching be performed on the list of audio annotations. In other
        cases (i.e., when `range_annotation` is None or Annotation), it defaults
        to a searching process on the list of text annotations.

        Args:
            entry_type: The type of entry to iterate over.
            range_annotation: The range annotation that serve as the constraint.

        Returns:
            An iterator of the entries with in the `range_annotation`.

        """
        use_coverage = self._index.coverage_index_is_valid
        coverage_index: Optional[Dict[int, Set[int]]] = {}

        if use_coverage:
            coverage_index = self._index.coverage_index(
                type(range_annotation), entry_type
            )
            if coverage_index is None:
                use_coverage = False

        def get_bisect_range(entry_class, search_list: SortedList):
            """
            Perform binary search on the specified list for target entry class.

            Args:
                entry_class: Target type of entry. It can be Annotation or
                    `AudioAnnotation`.
                search_list: A `SortedList` object on which the binary search
                    will be carried out.
            """
            range_begin = range_annotation.begin if range_annotation else 0
            range_end = (
                range_annotation.end
                if range_annotation
                else search_list[-1].end
            )

            temp_begin = entry_class(self, range_begin, range_begin)
            begin_index = search_list.bisect(temp_begin)

            temp_end = entry_class(self, range_end, range_end)
            end_index = search_list.bisect(temp_end)

            # Make sure these temporary annotations are not part of the
            # actual data.
            temp_begin.regret_creation()
            temp_end.regret_creation()
            return search_list[begin_index:end_index]

        if use_coverage and coverage_index is not None:
            for tid in coverage_index[range_annotation.tid]:
                yield self.get_entry(tid)  # type: ignore
        elif isinstance(range_annotation, AudioAnnotation):
            if issubclass(entry_type, AudioAnnotation):
                yield from get_bisect_range(
                    AudioAnnotation, self.audio_annotations
                )
            elif issubclass(entry_type, Link):
                for link in self.links:
                    if self._index.in_audio_span(link, range_annotation.span):
                        yield link
            elif issubclass(entry_type, Group):
                for group in self.groups:
                    if self._index.in_audio_span(group, range_annotation.span):
                        yield group
        else:
            if issubclass(entry_type, Annotation):
                yield from get_bisect_range(Annotation, self.annotations)
            elif issubclass(entry_type, Link):
                for link in self.links:
                    if self._index.in_span(link, range_annotation.span):
                        yield link
            elif issubclass(entry_type, Group):
                for group in self.groups:
                    if self._index.in_span(group, range_annotation.span):
                        yield group

    def get(  # type: ignore
        self,
        entry_type: Union[str, Type[EntryType]],
        range_annotation: Optional[Union[Annotation, AudioAnnotation]] = None,
        components: Optional[Union[str, Iterable[str]]] = None,
        include_sub_type: bool = True,
    ) -> Iterable[EntryType]:
        r"""This function is used to get data from a data pack with various
        methods.

        Depending on the provided arguments, the function will perform several
        different filtering of the returned data.

        The ``entry_type`` is mandatory, where all the entries matching this
        type
        will be returned. The sub-types of the provided entry type will be
        also returned if ``include_sub_type`` is set to True (which is the
        default behavior).

        The ``range_annotation`` controls the search area of the sub-types. An
        entry `E` will be returned if
        :meth:`~forte.data.data_pack.DataIndex.in_span` or
        :meth:`~forte.data.data_pack.DataIndex.in_audio_span` returns True.
        If this function is called frequently
        with queries related to the ``range_annotation``, please consider to
        build
        the coverage index regarding the related entry types. User can call
        :meth:`build_coverage_for(context_type, covered_type)` in order to
        build
        a mapping between a pair of entry types and target entries that are
        covered in ranges specified by outer entries.

        The ``components`` list will filter the results by the `component` (i.e
        the creator of the entry). If ``components`` is provided, only the
        entries
        created by one of the ``components`` will be returned.

        Example:

            .. code-block:: python

                # Iterate through all the sentences in the pack.
                for sentence in input_pack.get(Sentence):
                    # Take all tokens from a sentence created by NLTKTokenizer.
                    token_entries = input_pack.get(
                        entry_type=Token,
                        range_annotation=sentence,
                        component='NLTKTokenizer')
                    ...

            In the above code snippet, we get entries of type ``Token`` within
            each ``sentence`` which were generated by ``NLTKTokenizer``. You
            can consider build coverage index between ``Token`` and
            ``Sentence``
            if this snippet is frequently used:

                .. code-block:: python

                    # Build coverage index between `Token` and `Sentence`
                    input_pack.build_coverage_for(
                        context_type=Sentence
                        covered_type=Token
                    )

            After building the index from the snippet above, you will be able
            to retrieve the tokens covered by sentence much faster.


        Args:
            entry_type: The type of entries requested.
            range_annotation: The
                range of entries requested. If `None`, will return valid
                entries in the range of whole data pack.
            components: The component (creator)
                generating the entries requested. If `None`, will return valid
                entries generated by any component.
            include_sub_type: whether to consider the sub types of
                the provided entry type. Default `True`.

        Yields:
            Each `Entry` found using this method.
        """
        entry_type_: Type[EntryType] = as_entry_type(entry_type)

        def require_annotations(entry_class=Annotation) -> bool:
            if issubclass(entry_type_, entry_class):
                return True
            if issubclass(entry_type_, Link):
                return issubclass(
                    entry_type_.ParentType, entry_class
                ) and issubclass(entry_type_.ChildType, entry_class)
            if issubclass(entry_type_, Group):
                return issubclass(entry_type_.MemberType, entry_class)
            return False

        # If we don't have any annotations but the items to check requires them,
        # then we simply yield from an empty list.
        if (
            len(self.annotations) == 0
            and isinstance(range_annotation, Annotation)
            and require_annotations(Annotation)
        ) or (
            len(self.audio_annotations) == 0
            and isinstance(range_annotation, AudioAnnotation)
            and require_annotations(AudioAnnotation)
        ):
            yield from []
            return

        # If the ``entry_type`` and `range_annotation` are for different types of
        # payload, then we yield from an empty list with a warning.
        if (
            require_annotations(Annotation)
            and isinstance(range_annotation, AudioAnnotation)
        ) or (
            require_annotations(AudioAnnotation)
            and isinstance(range_annotation, Annotation)
        ):
            logger.warning(
                "Incompatible combination of ``entry_type`` and "
                "`range_annotation` found in the input of `DataPack.get()`"
                " method. An empty iterator will be returned when inputs "
                "contain multi-media entries. Please double check the input "
                "arguments and make sure they are associated with the same type"
                " of payload (i.e., either text or audio)."
            )
            yield from []
            return

        # Valid entry ids based on type.
        all_types: Set[Type]
        if include_sub_type:
            all_types = self._expand_to_sub_types(entry_type_)
        else:
            all_types = {entry_type_}

        entry_iter: Iterator[Entry]
        if issubclass(entry_type_, Generics):
            entry_iter = self.generics
        elif isinstance(range_annotation, (Annotation, AudioAnnotation)):
            if (
                issubclass(entry_type_, Annotation)
                or issubclass(entry_type_, Link)
                or issubclass(entry_type_, Group)
                or issubclass(entry_type_, AudioAnnotation)
            ):
                entry_iter = self.iter_in_range(entry_type_, range_annotation)
        elif issubclass(entry_type_, Annotation):
            entry_iter = self.annotations
        elif issubclass(entry_type_, Link):
            entry_iter = self.links
        elif issubclass(entry_type_, Group):
            entry_iter = self.groups
        elif issubclass(entry_type_, AudioAnnotation):
            entry_iter = self.audio_annotations
        else:
            raise ValueError(
                f"The requested type {str(entry_type_)} is not supported."
            )

        for entry in entry_iter:
            # Filter by type and components.
            if type(entry) not in all_types:
                continue
            if components is not None:
                if not self.is_created_by(entry, components):
                    continue
            yield entry  # type: ignore

    def update(self, datapack: "DataPack"):
        r"""Update the attributes and properties of the current DataPack with
        another DataPack.

        Args:
            datapack: A reference datapack to update
        """
        # TODO: Not recommended to directly update __dict__. Should find a
        #   better solution.
        self.__dict__.update(datapack.__dict__)


class DataIndex(BaseIndex):
    r"""A set of indexes used in :class:`~forte.data.data_pack.DataPack`, note that this class is
    used by the `DataPack` internally.

    #. :attr:`entry_index`, the index from each ``tid`` to the corresponding entry
    #. :attr:`type_index`, the index from each type to the entries of
       that type
    #. :attr:`component_index`, the index from each component to the
       entries generated by that component
    #. :attr:`link_index`, the index from child
       (:attr:`link_index["child_index"]`)and parent
       (:attr:`link_index["parent_index"]`) nodes to links
    #. :attr:`group_index`, the index from group members to groups.
    #. :attr:`_coverage_index`, the index that maps from an annotation to
       the entries it covers. :attr:`_coverage_index` is a dict of dict, where
       the key is a tuple of the outer entry type and the inner entry type.
       The outer entry type should be an annotation type. The value is a dict,
       where the key is the ``tid`` of the outer entry, and the value is a set of
       ``tid`` that are covered by the outer entry. We say an Annotation A covers
       an entry E if one of the following condition is met:
       1. E is of Annotation type, and that E.begin >= A.begin, E.end <= E.end
       2. E is of Link type, and both E's parent and child node are Annotation
       that are covered by A.

    """

    def __init__(self):
        super().__init__()
        self._coverage_index: Dict[
            Tuple[Type[Union[Annotation, AudioAnnotation]], Type[EntryType]],
            Dict[int, Set[int]],
        ] = {}
        self._coverage_index_valid = True

    def remove_entry(self, entry: EntryType):
        super().remove_entry(entry)
        self.deactivate_coverage_index()

    @property
    def coverage_index_is_valid(self):
        return self._coverage_index_valid

    def activate_coverage_index(self):
        self._coverage_index_valid = True

    def deactivate_coverage_index(self):
        self._coverage_index_valid = False

    def coverage_index(
        self,
        outer_type: Type[Union[Annotation, AudioAnnotation]],
        inner_type: Type[EntryType],
    ) -> Optional[Dict[int, Set[int]]]:
        r"""Get the coverage index from ``outer_type`` to ``inner_type``.

        Args:
            outer_type: an annotation or `AudioAnnotation` type.
            inner_type: an entry type.

        Returns:
            If the coverage index does not exist, return `None`. Otherwise,
            return a dict.
        """
        if not self.coverage_index_is_valid:
            return None
        return self._coverage_index.get((outer_type, inner_type))

    def get_covered(
        self,
        data_pack: DataPack,
        context_annotation: Union[Annotation, AudioAnnotation],
        inner_type: Type[EntryType],
    ) -> Set[int]:
        """
        Get the entries covered by a certain context annotation

        Args:
            data_pack: The data pack to search for.
            context_annotation: The context annotation to search in.
            inner_type: The inner type to be searched for.

        Returns:
            Entry ID of type `inner_type` that is covered by
            `context_annotation`.
        """
        context_type = context_annotation.__class__
        if self.coverage_index(context_type, inner_type) is None:
            self.build_coverage_index(data_pack, context_type, inner_type)
        assert self._coverage_index is not None
        return self._coverage_index.get((context_type, inner_type), {}).get(
            context_annotation.tid, set()
        )

    def build_coverage_index(
        self,
        data_pack: DataPack,
        outer_type: Type[Union[Annotation, AudioAnnotation]],
        inner_type: Type[EntryType],
    ):
        r"""Build the coverage index from ``outer_type`` to ``inner_type``.

        Args:
            data_pack: The data pack to build coverage for.
            outer_type: an annotation or `AudioAnnotation` type.
            inner_type: an entry type, can be Annotation, Link, Group,
                `AudioAnnotation`.
        """
        if not issubclass(
            inner_type, (Annotation, Link, Group, AudioAnnotation)
        ):
            raise ValueError(f"Do not support coverage index for {inner_type}.")

        if not self.coverage_index_is_valid:
            self._coverage_index = {}

        # prevent the index from being used during construction
        self.deactivate_coverage_index()

        # TODO: tests and documentations for the edge cases are missing. i.e. we
        #  are not clear about what would happen if the covered annotation
        #  is the same as the covering annotation, or if their spans are the
        #  same.
        self._coverage_index[(outer_type, inner_type)] = {}
        for range_annotation in data_pack.get_entries_of(outer_type):
            if isinstance(range_annotation, (Annotation, AudioAnnotation)):
                entries = data_pack.get(inner_type, range_annotation)
                entry_ids = {e.tid for e in entries}
                self._coverage_index[(outer_type, inner_type)][
                    range_annotation.tid
                ] = entry_ids

        self.activate_coverage_index()

    def have_overlap(
        self,
        entry1: Union[Annotation, int, AudioAnnotation],
        entry2: Union[Annotation, int, AudioAnnotation],
    ) -> bool:
        r"""Check whether the two annotations have overlap in span.

        Args:
            entry1: An
                :class:`Annotation` or :class:`AudioAnnotation` object to be
                checked, or the ``tid`` of the Annotation.
            entry2: Another
                :class:`Annotation` or :class:`AudioAnnotation` object to be
                checked, or the ``tid`` of the Annotation.
        """
        entry1_: Union[Annotation, AudioAnnotation] = (
            self._entry_index[entry1]
            if isinstance(entry1, (int, np.integer))
            else entry1
        )
        entry2_: Union[Annotation, AudioAnnotation] = (
            self._entry_index[entry2]
            if isinstance(entry2, (int, np.integer))
            else entry2
        )

        if not isinstance(entry1_, (Annotation, AudioAnnotation)):
            raise TypeError(
                f"'entry1' should be an instance of Annotation or `AudioAnnotation`,"
                f" but get {type(entry1)}"
            )

        if not isinstance(entry2_, (Annotation, AudioAnnotation)):
            raise TypeError(
                f"'entry2' should be an instance of Annotation or `AudioAnnotation`,"
                f" but get {type(entry2)}"
            )

        if (
            isinstance(entry1_, Annotation)
            and isinstance(entry2_, AudioAnnotation)
        ) or (
            isinstance(entry1_, AudioAnnotation)
            and isinstance(entry2_, Annotation)
        ):
            raise TypeError(
                "'entry1' and 'entry2' should be the same type of entry, "
                f"but get type(entry1)={type(entry1_)}, "
                f"typr(entry2)={type(entry2_)}"
            )

        return not (
            entry1_.begin >= entry2_.end or entry1_.end <= entry2_.begin
        )

    def in_span(self, inner_entry: Union[int, Entry], span: Span) -> bool:
        r"""Check whether the ``inner entry`` is within the given ``span``. The
        criterion are as followed:

        Annotation entries: they are considered in a span if the
        begin is not smaller than `span.begin` and the end is not larger than
        `span.end`.

        Link entries: if the parent and child of the links are both
        `Annotation` type, this link will be considered in span if both parent
        and child are :meth:`~forte.data.data_pack.DataIndex.in_span` of the
        provided `span`. If either the parent and
        the child is not of type `Annotation`, this function will always return
        `False`.

        Group entries: if the child type of the group is `Annotation` type,
        then the group will be considered in span if all the elements are
        :meth:`~forte.data.data_pack.DataIndex.in_span` of the provided `span`.
        If the child type is not `Annotation`
        type, this function will always return `False`.

        Other entries (i.e Generics and `AudioAnnotation`): they will not be
        considered :meth:`~forte.data.data_pack.DataIndex.in_span` of any
        spans. The function will always return
        `False`.

        Args:
            inner_entry: The inner entry object to be checked
             whether it is within ``span``. The argument can be the entry id
             or the entry object itself.
            span: A :class:`~forte.data.span.Span` object to be checked. We will check
                whether the ``inner_entry`` is within this span.

        Returns:
            True if the `inner_entry` is considered to be in span of the
            provided span.
        """
        # The reason of this check is that the get_data method will use numpy
        # integers. This might create problems when other unexpected integers
        # are used.
        if isinstance(inner_entry, (int, np.integer)):
            inner_entry = self._entry_index[inner_entry]

        inner_begin = -1
        inner_end = -1

        if isinstance(inner_entry, Annotation):
            inner_begin = inner_entry.begin
            inner_end = inner_entry.end
        elif isinstance(inner_entry, Link):
            if not issubclass(inner_entry.ParentType, Annotation):
                return False

            if not issubclass(inner_entry.ChildType, Annotation):
                return False

            child = inner_entry.get_child()
            parent = inner_entry.get_parent()

            if not isinstance(child, Annotation) or not isinstance(
                parent, Annotation
            ):
                # Cannot check in_span for non-annotations.
                return False

            child_: Annotation = child
            parent_: Annotation = parent

            inner_begin = min(child_.begin, parent_.begin)
            inner_end = max(child_.end, parent_.end)
        elif isinstance(inner_entry, Group):
            if not issubclass(inner_entry.MemberType, Annotation):
                return False

            for mem in inner_entry.get_members():
                mem_: Annotation = mem  # type: ignore
                if inner_begin == -1:
                    inner_begin = mem_.begin
                inner_begin = min(inner_begin, mem_.begin)
                inner_end = max(inner_end, mem_.end)
        else:
            # Generics, AudioAnnotation, or other user defined types will not
            # be check here.
            return False
        return inner_begin >= span.begin and inner_end <= span.end

    def in_audio_span(self, inner_entry: Union[int, Entry], span: Span) -> bool:
        r"""Check whether the ``inner entry`` is within the given audio span.
        This method is identical to
        :meth::meth:`~forte.data.data_pack.DataIndex.in_span` except that it
        operates on
        the audio payload of datapack. The criterion are as followed:

        `AudioAnnotation` entries: they are considered in a span if the
        begin is not smaller than `span.begin` and the end is not larger than
        `span.end`.

        Link entries: if the parent and child of the links are both
        `AudioAnnotation` type, this link will be considered in span if both
        parent and child are :meth:`~forte.data.data_pack.DataIndex.in_span` of
        the provided `span`. If either the
        parent and the child is not of type `AudioAnnotation`, this function
        will always return `False`.

        Group entries: if the child type of the group is `AudioAnnotation`
        type,
        then the group will be considered in span if all the elements are
        :meth:`~forte.data.data_pack.DataIndex.in_span` of the provided `span`.
        If the child type is not
        `AudioAnnotation` type, this function will always return `False`.

        Other entries (i.e Generics and Annotation): they will not be
        considered
        :meth:`~forte.data.data_pack.DataIndex.in_span` of any spans. The
        function will always return `False`.

        Args:
            inner_entry: The inner entry object to be checked
                whether it is within ``span``. The argument can be the entry id
                or the entry object itself.
            span: A :class:`~forte.data.span.Span` object to be checked.
                We will check whether the ``inner_entry`` is within this span.

        Returns:
            True if the `inner_entry` is considered to be in span of the
            provided span.
        """
        # The reason of this check is that the get_data method will use numpy
        # integers. This might create problems when other unexpected integers
        # are used.
        if isinstance(inner_entry, (int, np.integer)):
            inner_entry = self._entry_index[inner_entry]

        inner_begin = -1
        inner_end = -1

        if isinstance(inner_entry, AudioAnnotation):
            inner_begin = inner_entry.begin
            inner_end = inner_entry.end
        elif isinstance(inner_entry, Link):
            if not (
                issubclass(inner_entry.ParentType, AudioAnnotation)
                and issubclass(inner_entry.ChildType, AudioAnnotation)
            ):
                return False

            child = inner_entry.get_child()
            parent = inner_entry.get_parent()

            if not isinstance(child, AudioAnnotation) or not isinstance(
                parent, AudioAnnotation
            ):
                # Cannot check in_span for non-AudioAnnotation.
                return False

            child_: AudioAnnotation = child
            parent_: AudioAnnotation = parent

            inner_begin = min(child_.begin, parent_.begin)
            inner_end = max(child_.end, parent_.end)
        elif isinstance(inner_entry, Group):
            if not issubclass(inner_entry.MemberType, AudioAnnotation):
                return False

            for mem in inner_entry.get_members():
                mem_: AudioAnnotation = mem  # type: ignore
                if inner_begin == -1:
                    inner_begin = mem_.begin
                inner_begin = min(inner_begin, mem_.begin)
                inner_end = max(inner_end, mem_.end)
        else:
            # Generics, Annotation, or other user defined types will not be
            # check here.
            return False
        return inner_begin >= span.begin and inner_end <= span.end
